import struct
import unittest
from array import array
import struct
import platform

import moderngl

from common import get_context


class TestCase(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.ctx = get_context()

        prog = cls.ctx.program(
            vertex_shader='''
                #version 330
                in vec2 vert;
                out vec2 text;
                void main() {
                    gl_Position = vec4(vert * 2.0 - 1.0, 0.0, 1.0);
                    text = vert;
                }
            ''',
            fragment_shader='''
                #version 330
                uniform sampler2D Texture;
                in vec2 text;
                out vec4 color;
                void main() {
                    color = texture(Texture, text);
                }
            ''',
        )

        vbo = cls.ctx.buffer(struct.pack('8f', 0, 0, 0, 1, 1, 0, 1, 1))
        cls.vao = cls.ctx.simple_vertex_array(prog, vbo, 'vert')

    def test_1(self):
        self.assertEqual(self.vao.mode, moderngl.TRIANGLES)
        fbo = self.ctx.framebuffer(self.ctx.renderbuffer((16, 16)))
        pixels = struct.pack('16B', 255, 0, 0, 255, 0, 255, 0, 255, 0, 0, 255, 255, 0, 0, 0, 255)
        texture = self.ctx.texture((2, 2), 4, pixels)

        fbo.use()
        texture.use()

        texture.filter = (moderngl.LINEAR, moderngl.LINEAR)
        self.vao.render(moderngl.TRIANGLE_STRIP)

        r, g, b = struct.unpack('3B', fbo.read((8, 8, 1, 1)))
        self.assertTrue(abs(r - 49) < 8 and abs(g - 63) < 8 and abs(b - 63) < 8)

        texture.filter = (moderngl.NEAREST, moderngl.NEAREST)
        self.vao.render(moderngl.TRIANGLE_STRIP)

        r, g, b = struct.unpack('3B', fbo.read((6, 6, 1, 1)))
        self.assertTrue(abs(r - 255) < 8 and abs(g - 0) < 8 and abs(b - 0) < 8)

        r, g, b = struct.unpack('3B', fbo.read((9, 6, 1, 1)))
        self.assertTrue(abs(r - 0) < 8 and abs(g - 255) < 8 and abs(b - 0) < 8)

        r, g, b = struct.unpack('3B', fbo.read((6, 9, 1, 1)))
        self.assertTrue(abs(r - 0) < 8 and abs(g - 0) < 8 and abs(b - 255) < 8)

        r, g, b = struct.unpack('3B', fbo.read((9, 9, 1, 1)))
        self.assertTrue(abs(r - 0) < 8 and abs(g - 0) < 8 and abs(b - 0) < 8)

    def test_2(self):
        fbo = self.ctx.framebuffer(self.ctx.renderbuffer((16, 16)))
        pixels1 = struct.pack('16B', 255, 0, 0, 255, 0, 255, 0, 255, 255, 0, 0, 255, 0, 255, 0, 255)
        pixels2 = struct.pack('16B', 20, 30, 40, 255, 80, 70, 60, 255, 20, 30, 40, 255, 80, 70, 60, 255)
        texture = self.ctx.texture((2, 2), 4, pixels1)

        fbo.use()
        texture.use()

        texture.filter = (moderngl.NEAREST, moderngl.NEAREST)
        self.vao.render(moderngl.TRIANGLE_STRIP)

        r, g, b = struct.unpack('3B', fbo.read((5, 7, 1, 1)))
        self.assertTrue(abs(r - 255) < 8 and abs(g - 0) < 8 and abs(b - 0) < 8)

        r, g, b = struct.unpack('3B', fbo.read((10, 7, 1, 1)))
        self.assertTrue(abs(r - 0) < 8 and abs(g - 255) < 8 and abs(b - 0) < 8)

        texture.write(pixels2)

        texture.filter = (moderngl.NEAREST, moderngl.NEAREST)
        self.vao.render(moderngl.TRIANGLE_STRIP)

        r, g, b = struct.unpack('3B', fbo.read((5, 7, 1, 1)))
        self.assertTrue(abs(r - 20) < 8 and abs(g - 30) < 8 and abs(b - 40) < 8)

        r, g, b = struct.unpack('3B', fbo.read((10, 7, 1, 1)))
        self.assertTrue(abs(r - 80) < 8 and abs(g - 70) < 8 and abs(b - 60) < 8)

    def test_3(self):
        fbo = self.ctx.framebuffer(self.ctx.renderbuffer((16, 16)))
        pixels1 = struct.pack('16B', 255, 0, 0, 255, 0, 255, 0, 255, 255, 0, 0, 255, 0, 255, 0, 255)
        pixels2 = struct.pack('16B', 20, 30, 40, 255, 80, 70, 60, 255, 20, 30, 40, 255, 80, 70, 60, 255)
        texture = self.ctx.texture((2, 2), 4)

        pbo1 = self.ctx.buffer(pixels1)
        pbo2 = self.ctx.buffer(pixels2)

        fbo.use()
        texture.use()

        texture.write(pbo1)
        texture.filter = (moderngl.NEAREST, moderngl.NEAREST)
        self.vao.render(moderngl.TRIANGLE_STRIP)

        r, g, b = struct.unpack('3B', fbo.read((5, 7, 1, 1)))
        self.assertTrue(abs(r - 255) < 8 and abs(g - 0) < 8 and abs(b - 0) < 8)

        r, g, b = struct.unpack('3B', fbo.read((10, 7, 1, 1)))
        self.assertTrue(abs(r - 0) < 8 and abs(g - 255) < 8 and abs(b - 0) < 8)

        texture.write(pbo2)
        texture.filter = (moderngl.NEAREST, moderngl.NEAREST)
        self.vao.render(moderngl.TRIANGLE_STRIP)

        r, g, b = struct.unpack('3B', fbo.read((5, 7, 1, 1)))
        self.assertTrue(abs(r - 20) < 8 and abs(g - 30) < 8 and abs(b - 40) < 8)

        r, g, b = struct.unpack('3B', fbo.read((10, 7, 1, 1)))
        self.assertTrue(abs(r - 80) < 8 and abs(g - 70) < 8 and abs(b - 60) < 8)

    def test_4(self):
        pixels = struct.pack('16B', 255, 0, 0, 255, 0, 255, 0, 255, 255, 0, 0, 255, 0, 255, 0, 255)
        texture = self.ctx.texture((2, 2), 4, pixels)
        self.assertEqual(texture.read(), pixels)

    def test_5(self):
        pixels = struct.pack('16B', 255, 0, 0, 255, 0, 255, 0, 255, 255, 0, 0, 255, 0, 255, 0, 255)
        texture = self.ctx.texture((2, 2), 4, pixels)
        buf = bytearray(len(pixels))
        texture.read_into(buf)

        self.assertEqual(bytes(buf), pixels)

    def test_6(self):
        pixels = struct.pack('16B', 255, 0, 0, 255, 0, 255, 0, 255, 255, 0, 0, 255, 0, 255, 0, 255)
        texture = self.ctx.texture((2, 2), 4, pixels)
        pbo = self.ctx.buffer(reserve=len(pixels))
        texture.read_into(pbo)

        self.assertEqual(pbo.read(), pixels)

    def test_override_internalformat(self):
        """Ensure no errors occur when overriding internalformat"""
        self.ctx.error
        if not "GL_EXT_texture_sRGB" in self.ctx.extensions:
            self.skipTest('GL_EXT_texture_sRGB extension not supported')

        GL_SRGB8 = 0x8C41
        pixels = struct.pack('16B', 255, 0, 0, 255, 0, 255, 0, 255, 255, 0, 0, 255, 0, 255, 0, 255)
        texture = self.ctx.texture((2, 2), 4, pixels, internal_format=GL_SRGB8)
        data = texture.read()
        self.assertEqual(self.ctx.error, "GL_NO_ERROR")

    def test_normalized_textures(self):
        """8 and 16 bit normalized integer textures"""
        ni1 = self.ctx.texture((4, 4), 4, dtype="ni1")
        ni2 = self.ctx.texture((4, 4), 4, dtype="ni2")
        nu1 = self.ctx.texture((4, 4), 4, dtype="nu1")
        nu2 = self.ctx.texture((4, 4), 4, dtype="nu2")

        nu1_data = array('B', [255] * 16 * 4).tobytes()
        nu2_data = array('H', [65535] * 16 * 4).tobytes()
        ni1_data = array('b', [127] * 16 * 4).tobytes()
        ni2_data = array('h', [32767] * 16 * 4).tobytes()

        nu1.write(nu1_data)
        nu2.write(nu2_data)
        ni1.write(ni1_data)
        ni2.write(ni2_data)

        self.assertEqual(nu1.read(), nu1_data)
        self.assertEqual(nu2.read(), nu2_data)
        self.assertEqual(ni1.read(), ni1_data)
        self.assertEqual(ni2.read(), ni2_data)

        fbo = self.ctx.simple_framebuffer((4, 4))
        fbo.use()

        # Render these textures to an RGBA8 framebuffer and ensure the result is a pure white color (FF)
        fbo.clear()
        nu1.use()
        self.vao.render()
        self.assertEqual(fbo.read(viewport=(0, 0, 1, 1), components=4, dtype="f1"), b'\xff\xff\xff\xff')

        fbo.clear()
        nu2.use()
        self.vao.render()
        self.assertEqual(fbo.read(viewport=(0, 0, 1, 1), components=4, dtype="f1"), b'\xff\xff\xff\xff')

        fbo.clear()
        ni1.use()
        self.vao.render()
        self.assertEqual(fbo.read(viewport=(0, 0, 1, 1), components=4, dtype="f1"), b'\xff\xff\xff\xff')

        fbo.clear()
        ni2.use()
        self.vao.render()
        self.assertEqual(fbo.read(viewport=(0, 0, 1, 1), components=4, dtype="f1"), b'\xff\xff\xff\xff')

    def test_depth_texture_write(self):
        """Write data into depth texture"""
        size = 4, 4
        fbo = self.ctx.framebuffer(
            color_attachments=[self.ctx.texture(size, 4)],
            depth_attachment=self.ctx.depth_texture(size),
        )
        fbo.clear(depth=1.0)

        first = struct.unpack("16f", fbo.depth_attachment.read())
        second = (1.0,) * 16
        for a, b in zip(first, second):
            self.assertAlmostEqual(a, b, places=6)

        fbo.depth_attachment.write(struct.pack("16f", *([0.5] * 16)))
        first = struct.unpack("16f", fbo.depth_attachment.read())
        second = (0.5,) * 16
        for a, b in zip(first, second):
            self.assertAlmostEqual(a, b, places=5)


if __name__ == '__main__':
    unittest.main()



"""
Generates buffer data on the gpu using transform feedback.
We use an empty VertexArray running it vertices times outputting data to the buffer.

More fancy calculations can of course be done. This example just shows the concept.
"""
import struct
import numpy as np
from ported._example import Example


class GenerateData(Example):
    title = "Generate Data"
    gl_version = (3, 3)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                in float number;

                out float times_two;
                out float times_three;

                void main() {
                    times_two = number * 2.0;
                    times_three = number * 3.0;
                }
            ''',
            varyings=['times_two', 'times_three'],
            varyings_capture_mode='separate',
        )

        input_array = np.array([1.0, 2.0, 3.0, 4.0, 5.0], 'f4')
        self.input_buffer = self.ctx.buffer(input_array.tobytes())
        self.output_buffer_0 = self.ctx.buffer(reserve=self.input_buffer.size)
        self.output_buffer_1 = self.ctx.buffer(reserve=self.input_buffer.size)
        self.vao = self.ctx.vertex_array(self.prog, [(self.input_buffer, '1f', 'number')])
        self.vao.transform([self.output_buffer_0, self.output_buffer_1], vertices=len(input_array))

        output_array_0 = np.frombuffer(self.output_buffer_0.read(), 'f4')
        output_array_1 = np.frombuffer(self.output_buffer_1.read(), 'f4')
        print(output_array_0)
        print(output_array_1)

    def render(self, time, frame_time):
        exit(0)


if __name__ == '__main__':
    GenerateData.run()



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import struct

# import GLWindow
# import ModernGL

# # Window & Context

# wnd = GLWindow.create_window()
# ctx = ModernGL.create_context()

# vert = ctx.vertex_shader('''
#     #version 330

#     in vec3 vert;

#     uniform float znear;
#     uniform float zfar;
#     uniform float fovy;
#     uniform float ratio;

#     uniform vec3 center;
#     uniform vec3 eye;
#     uniform vec3 up;

#     mat4 perspective() {
#         float zmul = (-2.0 * znear * zfar) / (zfar - znear);
#         float ymul = 1.0 / tan(fovy * 3.14159265 / 360);
#         float xmul = ymul / ratio;

#         return mat4(
#             xmul, 0.0, 0.0, 0.0,
#             0.0, ymul, 0.0, 0.0,
#             0.0, 0.0, -1.0, -1.0,
#             0.0, 0.0, zmul, 0.0
#         );
#     }

#     mat4 lookat() {
#         vec3 forward = normalize(center - eye);
#         vec3 side = normalize(cross(forward, up));
#         vec3 upward = cross(side, forward);
#         return mat4(
#             side.x, upward.x, -forward.x, 0,
#             side.y, upward.y, -forward.y, 0,
#             side.z, upward.z, -forward.z, 0,
#             -dot(eye, side), -dot(eye, upward), dot(eye, forward), 1
#         );
#     }

#     void main() {
#         gl_Position = perspective() * lookat() * vec4(vert, 1.0);
#     }
# ''')

# frag = ctx.fragment_shader('''
#     #version 330

#     out vec4 color;

#     void main() {
#         color = vec4(0.04, 0.04, 0.04, 1.0);
#     }
# ''')

# width, height = wnd.size

# prog = ctx.program(vert, frag])

# prog.uniforms['znear'].value = 0.1
# prog.uniforms['zfar'].value = 1000.0
# prog.uniforms['ratio'].value = width / height
# prog.uniforms['fovy'].value = 60

# prog.uniforms['eye'].value = (3, 3, 3)
# prog.uniforms['center'].value = (0, 0, 0)
# prog.uniforms['up'].value = (0, 0, 1)

# grid = bytes()

# for i in range(0, 65):
#     grid += struct.pack('6f', i - 32, -32.0, 0.0, i - 32, 32.0, 0.0)
#     grid += struct.pack('6f', -32.0, i - 32, 0.0, 32.0, i - 32, 0.0)

# vbo = ctx.buffer(grid)
# vao = ctx.simple_vertex_array(prog, vbo, ['vert'])

# while wnd.update():
#     ctx.viewport = wnd.viewport
#     ctx.clear(0.9, 0.9, 0.9)
#     vao.render(ModernGL.LINES, 65 * 4)



import random
import math
import struct
import unittest

import moderngl
import numpy as np

from common import get_context


class TestBlend(unittest.TestCase):
    """
    Test blending and blend equations with float32 framebuffer.
    Emulate blend and equation values in python and compare with moderngl.
    """
    @classmethod
    def setUpClass(cls):
        """Set up program, vbo and fbo"""
        cls.ctx = get_context()
        cls.prog = cls.ctx.program(
            vertex_shader='''
                #version 330

                in vec2 in_vert;

                void main() {
                    gl_Position = vec4(in_vert, 0.0, 1.0);
                }
            ''',
            fragment_shader='''
                #version 330

                out vec4 fragColor;
                uniform vec4 color;

                void main() {
                    fragColor = color;
                }
            ''',
        )
        quad = [
            -1.0,  1.0,
            -1.0, -1.0,
            1.0, 1.0,
            1.0, -1.0,
        ]
        vbo = cls.ctx.buffer(np.array(quad, dtype='f4').tobytes())
        cls.vao = cls.ctx.simple_vertex_array(cls.prog, vbo, 'in_vert')
        cls.fbo = cls.ctx.framebuffer(color_attachments=[
            cls.ctx.texture((1, 1), 4, dtype='f4'),
        ])

    @classmethod
    def tearDownClass(cls):
        """Set everything back to default"""
        cls.ctx.disable(moderngl.BLEND)
        cls.ctx.blend_equation = moderngl.FUNC_ADD, moderngl.FUNC_ADD
        cls.ctx.blend_func = moderngl.SRC_ALPHA, moderngl.ONE_MINUS_SRC_ALPHA

    def unpack_fbo(self):
        return struct.unpack('4f', self.fbo.read(components=4, dtype='f4')[:16])

    def err(self):
        e = self.ctx.error
        if e != 'GL_NO_ERROR':
            raise ValueError(e)

    def cf(self, blend_func, val, src, dst, src_a, dst_a):
        """Calculate factor

        Args:
            blend_func: The blend func to use
            src: source value (single r, g, b, or a)
            dest: dest value (single r, g, b, or a)
            src_a: alpha value
            dst_a: alpha value
        """
        result = None
        if blend_func == moderngl.ZERO:
            result = 0
        if blend_func == moderngl.ONE:
            result = val
        if blend_func == moderngl.SRC_COLOR:
            result = val * src
        if blend_func == moderngl.ONE_MINUS_SRC_COLOR:
            result = val * (1.0 - src)
        if blend_func == moderngl.DST_COLOR:
            result = val * dst
        if blend_func == moderngl.ONE_MINUS_DST_COLOR:
            result = val * (1.0 - dst)
        if blend_func == moderngl.SRC_ALPHA:
            result = val * src_a
        if blend_func == moderngl.ONE_MINUS_SRC_ALPHA:
            result = val * (1.0 - src_a)
        if blend_func == moderngl.DST_ALPHA:
            result = val * dst_a
        if blend_func == moderngl.ONE_MINUS_DST_ALPHA:
            result = val * (1.0 - dst_a)

        if result is None:
            raise ValueError("Unsupported blend_func")

        # Don't clamp here. It needs to happen in the combine
        return result

    def bleq(self, bsx, bdx, sx, dx, func):
        """Blend equation. Note that MIN and MAX have special behavior.

        Args:
            bsx: blended source value
            bdx: blended dest value
            sx: original value from source
            dx: original value from dest
        """
        if func == moderngl.FUNC_ADD:
            return bsx + bdx
        elif func == moderngl.FUNC_SUBTRACT:
            return bsx - bdx
        elif func == moderngl.FUNC_REVERSE_SUBTRACT:
            return bdx - bsx
        elif func == moderngl.MIN:
            return min(sx, dx)
        elif func == moderngl.MAX:
            return max(sx, dx)
        else:
            raise ValueError('Unsupported blend equation: {}'.format(func))

    def blend_emulate(self, src, dst, blend_func, blend_equation):
        """Emulates blending in python"""
        sr = self.cf(blend_func[0], src[0], src[0], dst[0], src[3], dst[3])
        sg = self.cf(blend_func[0], src[1], src[1], dst[1], src[3], dst[3])
        sb = self.cf(blend_func[0], src[2], src[2], dst[2], src[3], dst[3])
        sa = self.cf(blend_func[2], src[3], src[3], dst[3], src[3], dst[3])

        dr = self.cf(blend_func[1], dst[0], src[0], dst[0], src[3], dst[3])
        dg = self.cf(blend_func[1], dst[1], src[1], dst[1], src[3], dst[3])
        db = self.cf(blend_func[1], dst[2], src[2], dst[2], src[3], dst[3])
        da = self.cf(blend_func[3], dst[3], src[3], dst[3], src[3], dst[3])

        return (
            self.bleq(sr, dr, src[0], dst[0], blend_equation[0]),
            self.bleq(sg, dg, src[1], dst[1], blend_equation[0]),
            self.bleq(sb, db, src[2], dst[2], blend_equation[0]),
            self.bleq(sa, da, src[3], dst[3], blend_equation[1]),
        )

    def blend_moderngl(self, src, dst, blend_func, blend_equation):
        """Performs blending in moderngl.

        Args:
            src (tuple): The source fragment (r, g, b, a)
            dst (tuple): The destination fragment (r, g, b, a)
            blend_func (tuple): The blend func
            blend_func (tuple): The blend equation
        Returns:
            The result fragment
        """
        self.fbo.use()
        self.fbo.clear(red=dst[0], green=dst[1], blue=dst[2], alpha=dst[3])
        self.ctx.enable(moderngl.BLEND)

        self.prog['color'].value = src
        self.ctx.blend_func = blend_func
        self.ctx.blend_equation = blend_equation
        self.vao.render(mode=moderngl.TRIANGLE_STRIP)
        return self.unpack_fbo()

    def assertEqualColors(self, a, b):
        self.assertAlmostEqual(a[0], b[0], places=2)
        self.assertAlmostEqual(a[1], b[1], places=2)
        self.assertAlmostEqual(a[2], b[2], places=2)
        self.assertAlmostEqual(a[3], b[3], places=2)

    def test_blend_default(self):
        src = (0.0, 1.0, 0.0, 0.2)
        dst = (1.0, 0.0, 1.0, 0.3)
        blend_func = (
            moderngl.SRC_ALPHA, moderngl.ONE_MINUS_SRC_ALPHA,
            moderngl.SRC_ALPHA, moderngl.ONE_MINUS_SRC_ALPHA,
        )
        blend_equation = moderngl.FUNC_ADD, moderngl.FUNC_ADD

        a = self.blend_emulate(src, dst, blend_func, blend_equation)
        b = self.blend_moderngl(src, dst, blend_func, blend_equation)
        self.assertEqualColors(a, b)

    def test_blend_separate(self):
        src = (0.2, 0.6, 0.1, 0.3)
        dst = (0.3, 0.1, 0.4, 0.2)
        blend_func = (
            moderngl.SRC_ALPHA, moderngl.ONE_MINUS_SRC_ALPHA,
            moderngl.ONE, moderngl.ONE,
        )
        blend_equation = moderngl.FUNC_ADD, moderngl.FUNC_ADD

        a = self.blend_emulate(src, dst, blend_func, blend_equation)
        b = self.blend_moderngl(src, dst, blend_func, blend_equation)
        self.assertEqualColors(a, b)

    def test_blend_separate_equation(self):
        src = (0.2, 0.6, 0.1, 0.3)
        dst = (0.3, 0.1, 0.4, 0.2)
        blend_func = (
            moderngl.SRC_ALPHA, moderngl.ONE_MINUS_SRC_ALPHA,
            moderngl.ONE, moderngl.ONE,
        )
        blend_equation = moderngl.FUNC_REVERSE_SUBTRACT, moderngl.MAX

        a = self.blend_emulate(src, dst, blend_func, blend_equation)
        b = self.blend_moderngl(src, dst, blend_func, blend_equation)
        self.assertEqualColors(a, b)

    def test_blend_chaos(self):
        """Pick N random colors with random blend modes and equations"""
        bf = [
            moderngl.ZERO, moderngl.ONE, moderngl.SRC_COLOR, moderngl.ONE_MINUS_SRC_COLOR,
            moderngl.DST_COLOR, moderngl.ONE_MINUS_DST_COLOR, moderngl.SRC_ALPHA,
            moderngl.ONE_MINUS_SRC_ALPHA, moderngl.DST_ALPHA, moderngl.ONE_MINUS_DST_ALPHA,
        ]
        eq = [
            moderngl.FUNC_ADD, moderngl.FUNC_SUBTRACT, moderngl.FUNC_REVERSE_SUBTRACT,
            moderngl.MIN, moderngl.MAX,
        ]

        for _ in range(1000):
            src = tuple([random.uniform(0, 1) for i in range(4)])
            dst = tuple([random.uniform(0, 1) for i in range(4)])
            blend_func = tuple([random.choice(bf) for i in range(4)])
            blend_equation = random.choice(eq), random.choice(eq)
            a = self.blend_emulate(src, dst, blend_func, blend_equation)
            b = self.blend_moderngl(src, dst, blend_func, blend_equation)
            self.assertEqualColors(a, b)

    def test_invalid_blend_func(self):
        # TypeError: Not iterable
        with self.assertRaises(TypeError):
            self.ctx.blend_func = moderngl.ONE

        # Incorrect tuple size
        with self.assertRaises(moderngl.error.Error):
            self.ctx.blend_func = moderngl.ONE,

        # TypeError: "Test" is not an integer
        with self.assertRaises(TypeError):
            self.ctx.blend_func = moderngl.ONE, "Test"

        # Incorrect tuple size=3
        with self.assertRaises(moderngl.error.Error):
            self.ctx.blend_func = moderngl.ONE, moderngl.ONE, moderngl.ONE

        # Incorrect tuple size=5
        with self.assertRaises(moderngl.error.Error):
            self.ctx.blend_func = (
                moderngl.ONE,
                moderngl.ONE,
                moderngl.ONE,
                moderngl.ONE,
                moderngl.ONE,
            )

        # Working versions
        self.ctx.blend_func = moderngl.SRC_ALPHA, moderngl.ONE_MINUS_SRC_ALPHA
        self.ctx.blend_func = (
            moderngl.SRC_ALPHA, moderngl.ONE_MINUS_SRC_ALPHA,
            moderngl.SRC_ALPHA, moderngl.ONE_MINUS_SRC_ALPHA,
        )

    def test_invalid_blend_equation(self):
        with self.assertRaises(TypeError):
            self.ctx.blend_equation = None

        with self.assertRaises(moderngl.error.Error):
            self.ctx.blend_equation = moderngl.MAX, moderngl.MAX, moderngl.MAX

        self.ctx.blend_equation = moderngl.MAX
        self.ctx.blend_equation = moderngl.MAX, moderngl.MAX

    def test_get_values(self):
        with self.assertRaises(NotImplementedError):
            self.ctx.blend_func

        with self.assertRaises(NotImplementedError):
            self.ctx.blend_equation



from typing import Any, Optional, Tuple, Union

from _moderngl import InvalidObject

from .buffer import Buffer

__all__ = ['TextureCube']


class TextureCube:
    """
    Cubemaps are a texture using the type GL_TEXTURE_CUBE_MAP.

    They are similar to 2D textures in that they have two dimensions.
    However, each mipmap level has 6 faces, with each face having the
    same size as the other faces.

    The width and height of a cubemap must be the same (ie: cubemaps are squares),
    but these sizes need not be powers of two.

    .. Note:: ModernGL enables ``GL_TEXTURE_CUBE_MAP_SEAMLESS`` globally
                to ensure filtering will be done across the cube faces.

    A Texture3D object cannot be instantiated directly, it requires a context.
    Use :py:meth:`Context.texture_cube` to create one.
    """

    __slots__ = ['mglo', '_size', '_components', '_dtype', '_glo', 'ctx', 'extra']

    def __init__(self):
        self.mglo = None  #: Internal representation for debug purposes only.
        self._size = (None, None)
        self._components = None
        self._dtype = None
        self._glo = None
        self.ctx = None  #: The context this object belongs to
        self.extra = None  #: Any - Attribute for storing user defined objects
        raise TypeError()

    def __repr__(self):
        if hasattr(self, '_glo'):
            return f"<{self.__class__.__name__}: {self._glo}>"
        else:
            return f"<{self.__class__.__name__}: INCOMPLETE>"

    def __eq__(self, other: Any):
        return type(self) is type(other) and self.mglo is other.mglo

    def __hash__(self) -> int:
        return id(self)

    def __del__(self) -> None:
        if not hasattr(self, "ctx"):
            return

        if self.ctx.gc_mode == "auto":
            self.release()
        elif self.ctx.gc_mode == "context_gc":
            self.ctx.objects.append(self.mglo)

    @property
    def size(self) -> Tuple[int, int]:
        """tuple: The size of the texture cube (single face)."""
        return self._size

    @property
    def components(self) -> int:
        """int: The number of components of the texture."""
        return self._components

    @property
    def dtype(self) -> str:
        """str: Data type."""
        return self._dtype

    @property
    def filter(self) -> Tuple[int, int]:
        """
        tuple: The minification and magnification filter for the texture.

        (Default ``(moderngl.LINEAR. moderngl.LINEAR)``)

        Example::

            texture.filter == (moderngl.NEAREST, moderngl.NEAREST)
            texture.filter == (moderngl.LINEAR_MIPMAP_LINEAR, moderngl.LINEAR)
            texture.filter == (moderngl.NEAREST_MIPMAP_LINEAR, moderngl.NEAREST)
            texture.filter == (moderngl.LINEAR_MIPMAP_NEAREST, moderngl.NEAREST)
        """
        return self.mglo.filter

    @filter.setter
    def filter(self, value: Tuple[int, int]) -> None:
        self.mglo.filter = value

    @property
    def swizzle(self) -> str:
        """
        str: The swizzle mask of the texture (Default ``'RGBA'``).

        The swizzle mask change/reorder the ``vec4`` value returned by the ``texture()`` function
        in a GLSL shaders. This is represented by a 4 character string were each
        character can be::

            'R' GL_RED
            'G' GL_GREEN
            'B' GL_BLUE
            'A' GL_ALPHA
            '0' GL_ZERO
            '1' GL_ONE

        Example::

            # Alpha channel will always return 1.0
            texture.swizzle = 'RGB1'

            # Only return the red component. The rest is masked to 0.0
            texture.swizzle = 'R000'

            # Reverse the components
            texture.swizzle = 'ABGR'
        """
        return self.mglo.swizzle

    @swizzle.setter
    def swizzle(self, value: str) -> None:
        self.mglo.swizzle = value

    @property
    def anisotropy(self) -> float:
        """
        float: Number of samples for anisotropic filtering (Default ``1.0``).

        The value will be clamped in range ``1.0`` and ``ctx.max_anisotropy``.

        Any value greater than 1.0 counts as a use of anisotropic filtering::

            # Disable anisotropic filtering
            texture.anisotropy = 1.0

            # Enable anisotropic filtering suggesting 16 samples as a maximum
            texture.anisotropy = 16.0
        """
        return self.mglo.anisotropy

    @anisotropy.setter
    def anisotropy(self, value: float) -> None:
        self.mglo.anisotropy = value

    @property
    def glo(self) -> int:
        """
        int: The internal OpenGL object.

        This values is provided for debug purposes only.
        """
        return self._glo

    def read(self, face: int, *, alignment: int = 1) -> bytes:
        """
        Read a face from the cubemap as bytes into system memory.

        Face values are::

            0: Positive X
            1: Negative X
            2: Positive Y
            3: Negative Y
            4: Positive Z
            5: Negative Z

        Args:
            face (int): The face to read.

        Keyword Args:
            alignment (int): The byte alignment of the pixels.
        """
        return self.mglo.read(face, alignment)

    def read_into(
        self,
        buffer: Any,
        face: int,
        *,
        alignment: int = 1,
        write_offset: int = 0,
    ) -> None:
        """
        Read a face from the cubemap texture.

        Read a face of the cubemap into a bytearray or :py:class:`~moderngl.Buffer`.
        The advantage of reading into a :py:class:`~moderngl.Buffer` is that pixel data
        does not need to travel all the way to system memory::

            # Reading pixel data into a bytearray
            data = bytearray(4)
            texture = ctx.texture_cube((2, 2), 1)
            texture.read_into(data, 0)

            # Reading pixel data into a buffer
            data = ctx.buffer(reserve=4)
            texture = ctx.texture_cube((2, 2), 1)
            texture.read_into(data, 0)

        Args:
            buffer (bytearray): The buffer that will receive the pixels.
            face (int): The face to read.

        Keyword Args:
            alignment (int): The byte alignment of the pixels.
            write_offset (int): The write offset.
        """
        if type(buffer) is Buffer:
            buffer = buffer.mglo

        return self.mglo.read_into(buffer, face, alignment, write_offset)

    def write(
        self,
        face: int,
        data: Any,
        viewport: Optional[Union[Tuple[int, int], Tuple[int, int, int, int]]] = None,
        *,
        alignment: int = 1,
    ) -> None:
        r"""
        Update the content of the texture.

        Update the content of a face in the cubemap from byte data
        or a moderngl :py:class:`~moderngl.Buffer`::

            # Write data from a moderngl Buffer
            data = ctx.buffer(reserve=4)
            texture = ctx.texture_cube((2, 2), 1)
            texture.write(0, data)

            # Write data from bytes
            data = b'\xff\xff\xff\xff'
            texture = ctx.texture_cube((2, 2), 1)
            texture.write(0, data)

        Args:
            face (int): The face to update.
            data (bytes): The pixel data.
            viewport (tuple): The viewport.

        Keyword Args:
            alignment (int): The byte alignment of the pixels.
        """
        if type(data) is Buffer:
            data = data.mglo

        self.mglo.write(face, data, viewport, alignment)

    def use(self, location: int = 0) -> None:
        """
        Bind the texture to a texture unit.

        The location is the texture unit we want to bind the texture.
        This should correspond with the value of the ``samplerCube``
        uniform in the shader because samplers read from the texture
        unit we assign to them::

            # Define what texture unit our two samplerCube uniforms should represent
            program['texture_a'] = 0
            program['texture_b'] = 1
            # Bind textures to the texture units
            first_texture.use(location=0)
            second_texture.use(location=1)

        Args:
            location (int): The texture location/unit.
        """
        self.mglo.use(location)

    def bind_to_image(self, unit: int, read: bool = True, write: bool = True, level: int = 0, format: int = 0) -> None:
        """
        Bind a texture to an image unit (OpenGL 4.2 required).

        This is used to bind textures to image units for shaders.
        The idea with image load/store is that the user can bind
        one of the images in a Texture to a number of image binding points
        (which are separate from texture image units). Shaders can read
        information from these images and write information to them,
        in ways that they cannot with textures.

        It's important to specify the right access type for the image.
        This can be set with the ``read`` and ``write`` arguments.
        Allowed combinations are:

        - **Read-only**: ``read=True`` and ``write=False``
        - **Write-only**: ``read=False`` and ``write=True``
        - **Read-write**: ``read=True`` and ``write=True``

        ``format`` specifies the format that is to be used when performing
        formatted stores into the image from shaders. ``format`` must be
        compatible with the texture's internal format. **By default the format
        of the texture is passed in. The format parameter is only needed
        when overriding this behavior.**

        Note that we bind the texture cube as layered to make
        all the faces accessible. This can be updated to map
        single faces in the future. The Z component in imageLoad/Store
        represents the face id we are writing to (0-5).

        More information:

        - https://www.khronos.org/opengl/wiki/Image_Load_Store
        - https://www.khronos.org/registry/OpenGL-Refpages/gl4/html/glBindImageTexture.xhtml

        Args:
            unit (int): Specifies the index of the image unit to which to bind the texture
            texture (:py:class:`moderngl.Texture`): The texture to bind
        Keyword Args:
            read (bool): Allows the shader to read the image (default: ``True``)
            write (bool): Allows the shader to write to the image (default: ``True``)
            level (int): Level of the texture to bind (default: ``0``).
            format (int): (optional) The OpenGL enum value representing the format (defaults to the texture's format)
        """
        self.mglo.bind(unit, read, write, level, format)

    def release(self) -> None:
        """Release the ModernGL object."""
        if self.mglo is not None:
            self.mglo.release()
            self.mglo = InvalidObject()



import _example


class Example(_example.Example):
    title = "Simple 2D Example"

    STATIC_COLOR_MODE = 0
    ATTRIB_COLOR_MODE = 1
    TEXTURE_MODE = 2

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                in vec2 in_vert;
                in vec3 in_color;
                in vec2 in_text;

                out vec3 v_color;
                out vec2 v_text;

                void main() {
                    gl_Position = vec4(in_vert, 0.0, 1.0);
                    v_color = in_color;
                    v_text = in_text;
                }
            ''',
            fragment_shader='''
                #version 330

                uniform sampler2D Texture;
                uniform int RenderMode;
                uniform vec3 Color;

                in vec3 v_color;
                in vec2 v_text;

                out vec4 f_color;

                void main() {
                    if (RenderMode == 0) {
                        f_color = vec4(Color, 1.0);
                    } else if (RenderMode == 1) {
                        f_color = vec4(v_color, 1.0);
                    } else if (RenderMode == 2) {
                        f_color = texture(Texture, v_text);
                    }
                }
            ''',
        )



import warnings
from collections import deque
from typing import Any, Deque, Dict, List, Optional, Set, Tuple, Union

from _moderngl import InvalidObject

from .buffer import Buffer
from .compute_shader import ComputeShader
from .conditional_render import ConditionalRender
from .framebuffer import Framebuffer
from .program import Program, detect_format
from .program_members import (  # noqa: F401
    Attribute,
    Subroutine,
    Uniform,
    UniformBlock,
    Varying,
)
from .query import Query
from .renderbuffer import Renderbuffer
from .sampler import Sampler
from .scope import Scope
from .texture import Texture
from .texture_3d import Texture3D
from .texture_array import TextureArray
from .texture_cube import TextureCube
from .vertex_array import VertexArray

try:
    from moderngl import mgl  # type: ignore
except ImportError:
    pass


__all__ = ['Context', 'create_context', 'create_standalone_context',
           'NOTHING', 'BLEND', 'DEPTH_TEST', 'CULL_FACE', 'RASTERIZER_DISCARD', 'PROGRAM_POINT_SIZE',
           'ZERO', 'ONE', 'SRC_COLOR', 'ONE_MINUS_SRC_COLOR', 'SRC_ALPHA', 'ONE_MINUS_SRC_ALPHA',
           'DST_ALPHA', 'ONE_MINUS_DST_ALPHA', 'DST_COLOR', 'ONE_MINUS_DST_COLOR',
           'FUNC_ADD', 'FUNC_SUBTRACT', 'FUNC_REVERSE_SUBTRACT', 'MIN', 'MAX',
           'DEFAULT_BLENDING', 'ADDITIVE_BLENDING', 'PREMULTIPLIED_ALPHA',
           'FIRST_VERTEX_CONVENTION', 'LAST_VERTEX_CONVENTION']

# Context Flags

#: Represents no states. Can be used with :py:meth:`Context.enable_only` to disable all states.
NOTHING = 0
#: Enable/disable blending
BLEND = 1
#: Enable/disable depth testing
DEPTH_TEST = 2
#: Enable/disable face culling
CULL_FACE = 4
#: Enable/disable rasterization
RASTERIZER_DISCARD = 8
#: Context flag: Enables ``gl_PointSize`` in vertex or geometry shaders.
#:
#: When enabled we can write to ``gl_PointSize`` in the vertex shader to specify the point size
#: for each individual point.
#:
#: If this value is not set in the shader the behavior is undefined. This means the points may
#: or may not appear depending if the drivers enforce some default value for ``gl_PointSize``.
#:
#: When disabled :py:attr:`Context.point_size` is used.
PROGRAM_POINT_SIZE = 16

# Blend functions
#: (0,0,0,0)
ZERO = 0x0000
#: (1,1,1,1)
ONE = 0x0001
#: (Rs0/kR,Gs0/kG,Bs0/kB,As0/kA)
SRC_COLOR = 0x0300
#: (1,1,1,1) − (Rs0/kR,Gs0/kG,Bs0/kB,As0/kA)
ONE_MINUS_SRC_COLOR = 0x0301
#: (As0/kA,As0/kA,As0/kA,As0/kA)
SRC_ALPHA = 0x0302
#: (1,1,1,1) − (As0/kA,As0/kA,As0/kA,As0/kA)
ONE_MINUS_SRC_ALPHA = 0x0303
#: (Ad/kA,Ad/kA,Ad/kA,Ad/kA)
DST_ALPHA = 0x0304
#: (1,1,1,1) − (Ad/kA,Ad/kA,Ad/kA,Ad/kA)
ONE_MINUS_DST_ALPHA = 0x0305
#: (Rd/kR,Gd/kG,Bd/kB,Ad/kA)
DST_COLOR = 0x0306
#: (1,1,1,1) − (Rd/kR,Gd/kG,Bd/kB,Ad/kA)
ONE_MINUS_DST_COLOR = 0x0307

# Blend equations
#: source + destination
FUNC_ADD = 0x8006
#: source - destination
FUNC_SUBTRACT = 0x800A
#: destination - source
FUNC_REVERSE_SUBTRACT = 0x800B
#: Minimum of source and destination
MIN = 0x8007
#: Maximum of source and destination
MAX = 0x8008

#: Shotcut for the default blending ``SRC_ALPHA, ONE_MINUS_SRC_ALPHA``
DEFAULT_BLENDING = (SRC_ALPHA, ONE_MINUS_SRC_ALPHA)
#: Shotcut for additive blending ``ONE, ONE``
ADDITIVE_BLENDING = (ONE, ONE)
#: Shotcut for blend mode when using premultiplied alpha ``SRC_ALPHA, ONE``
PREMULTIPLIED_ALPHA = (SRC_ALPHA, ONE)

#: Specifies the first vertex should be used as the source of data for flat shaded varyings.
#: Used with :py:attr:`Context.provoking_vertex`.
FIRST_VERTEX_CONVENTION = 0x8E4D
#: Specifies the last vertex should be used as the source of data for flat shaded varyings.
#: Used with :py:attr:`Context.provoking_vertex`.
LAST_VERTEX_CONVENTION = 0x8E4E


class Context:
    """
    Class exposing OpenGL features.

    ModernGL objects can be created from this class.
    """

    _valid_gc_modes = [None, "context_gc", "auto"]
    # Context Flags
    #: Represents no states. Can be used with :py:meth:`Context.enable_only` to disable all states.
    NOTHING = 0
    #: Enable/disable blending
    BLEND = 1
    #: Enable/disable depth testing
    DEPTH_TEST = 2
    #: Enable/disable face culling
    CULL_FACE = 4
    #: Enable/disable rasterization
    RASTERIZER_DISCARD = 8
    #: Context flag: Enables ``gl_PointSize`` in vertex or geometry shaders.
    #:
    #: When enabled we can write to ``gl_PointSize`` in the vertex shader to specify the point size
    #: for each individual point.
    #:
    #: If this value is not set in the shader the behavior is undefined. This means the points may
    #: or may not appear depending if the drivers enforce some default value for ``gl_PointSize``.
    #:
    #: When disabled :py:attr:`Context.point_size` is used.
    PROGRAM_POINT_SIZE = 16

    # Primitive modes

    #: Each vertex represents a point
    POINTS = 0x0000
    #: Vertices 0 and 1 are considered a line. Vertices 2 and 3 are considered a line.
    #: And so on. If the user specifies a non-even number of vertices, then the extra vertex is ignored.
    LINES = 0x0001
    #: As line strips, except that the first and last vertices are also used as a line.
    #: Thus, you get n lines for n input vertices. If the user only specifies 1 vertex,
    #: the drawing command is ignored. The line between the first and last vertices happens
    #: after all of the previous lines in the sequence.
    LINE_LOOP = 0x0002
    #: The adjacent vertices are considered lines. Thus, if you pass n vertices, you will get n-1 lines.
    #: If the user only specifies 1 vertex, the drawing command is ignored.
    LINE_STRIP = 0x0003
    #: Vertices 0, 1, and 2 form a triangle. Vertices 3, 4, and 5 form a triangle. And so on.
    TRIANGLES = 0x0004
    #: Every group of 3 adjacent vertices forms a triangle. The face direction of the
    #: strip is determined by the winding of the first triangle. Each successive triangle
    #: will have its effective face order reversed, so the system compensates for that
    #: by testing it in the opposite way. A vertex stream of n length will generate n-2 triangles.
    TRIANGLE_STRIP = 0x0005
    #: The first vertex is always held fixed. From there on, every group of 2 adjacent
    #: vertices form a triangle with the first. So with a vertex stream, you get a list
    #: of triangles like so: (0, 1, 2) (0, 2, 3), (0, 3, 4), etc. A vertex stream of
    #: n length will generate n-2 triangles.
    TRIANGLE_FAN = 0x0006
    #: These are special primitives that are expected to be used specifically with
    #: geomtry shaders. These primitives give the geometry shader more vertices
    #: to work with for each input primitive. Data needs to be duplicated in buffers.
    LINES_ADJACENCY = 0x000A
    #: These are special primitives that are expected to be used specifically with
    #: geomtry shaders. These primitives give the geometry shader more vertices
    #: to work with for each input primitive. Data needs to be duplicated in buffers.
    LINE_STRIP_ADJACENCY = 0x000B
    #: These are special primitives that are expected to be used specifically with
    #: geomtry shaders. These primitives give the geometry shader more vertices
    #: to work with for each input primitive. Data needs to be duplicated in buffers.
    TRIANGLES_ADJACENCY = 0x000C
    #: These are special primitives that are expected to be used specifically with
    #: geomtry shaders. These primitives give the geometry shader more vertices
    #: to work with for each input primitive. Data needs to be duplicated in buffers.
    TRIANGLE_STRIP_ADJACENCY = 0x0000D
    #: primitive type can only be used when Tessellation is active. It is a primitive
    #: with a user-defined number of vertices, which is then tessellated based on the
    #: control and evaluation shaders into regular points, lines, or triangles, depending
    #: on the TES's settings.
    PATCHES = 0x000E

    # Texture filters
    #: Returns the value of the texture element that is nearest
    #: (in Manhattan distance) to the specified texture coordinates.
    NEAREST = 0x2600
    #: Returns the weighted average of the four texture elements
    #: that are closest to the specified texture coordinates.
    #: These can include items wrapped or repeated from other parts
    #: of a texture, depending on the values of texture repeat mode,
    #: and on the exact mapping.
    LINEAR = 0x2601
    #: Chooses the mipmap that most closely matches the size of the
    #: pixel being textured and uses the ``NEAREST`` criterion (the texture
    #: element closest to the specified texture coordinates) to produce
    #: a texture value.
    NEAREST_MIPMAP_NEAREST = 0x2700
    #: Chooses the mipmap that most closely matches the size of the pixel
    #: being textured and uses the ``LINEAR`` criterion (a weighted average
    #: of the four texture elements that are closest to the specified
    #: texture coordinates) to produce a texture value.
    LINEAR_MIPMAP_NEAREST = 0x2701
    #: Chooses the two mipmaps that most closely match the size of the
    #: pixel being textured and uses the ``NEAREST`` criterion (the texture
    #: element closest to the specified texture coordinates ) to produce
    #: a texture value from each mipmap. The final texture value is a
    #: weighted average of those two values.
    NEAREST_MIPMAP_LINEAR = 0x2702
    #: Chooses the two mipmaps that most closely match the size of the pixel
    #: being textured and uses the ``LINEAR`` criterion (a weighted average
    #: of the texture elements that are closest to the specified texture
    #: coordinates) to produce a texture value from each mipmap.
    #: The final texture value is a weighted average of those two values.
    LINEAR_MIPMAP_LINEAR = 0x2703

    # Blend functions
    #: (0,0,0,0)
    ZERO = 0x0000
    #: (1,1,1,1)
    ONE = 0x0001
    #: (Rs0/kR,Gs0/kG,Bs0/kB,As0/kA)
    SRC_COLOR = 0x0300
    #: (1,1,1,1) − (Rs0/kR,Gs0/kG,Bs0/kB,As0/kA)
    ONE_MINUS_SRC_COLOR = 0x0301
    #: (As0/kA,As0/kA,As0/kA,As0/kA)
    SRC_ALPHA = 0x0302
    #: (1,1,1,1) − (As0/kA,As0/kA,As0/kA,As0/kA)
    ONE_MINUS_SRC_ALPHA = 0x0303
    #: (Ad/kA,Ad/kA,Ad/kA,Ad/kA)
    DST_ALPHA = 0x0304
    #: (1,1,1,1) − (Ad/kA,Ad/kA,Ad/kA,Ad/kA)
    ONE_MINUS_DST_ALPHA = 0x0305
    #: (Rd/kR,Gd/kG,Bd/kB,Ad/kA)
    DST_COLOR = 0x0306
    #: (1,1,1,1) − (Rd/kR,Gd/kG,Bd/kB,Ad/kA)
    ONE_MINUS_DST_COLOR = 0x0307

    #: Shotcut for the default blending ``SRC_ALPHA, ONE_MINUS_SRC_ALPHA``
    DEFAULT_BLENDING = (SRC_ALPHA, ONE_MINUS_SRC_ALPHA)
    #: Shotcut for additive blending ``ONE, ONE``
    ADDITIVE_BLENDING = (ONE, ONE)
    #: Shotcut for blend mode when using premultiplied alpha ``SRC_ALPHA, ONE``
    PREMULTIPLIED_ALPHA = (SRC_ALPHA, ONE)

    # Blend equations
    #: source + destination
    FUNC_ADD = 0x8006
    #: source - destination
    FUNC_SUBTRACT = 0x800A
    #: destination - source
    FUNC_REVERSE_SUBTRACT = 0x800B
    #: Minimum of source and destination
    MIN = 0x8007
    #: Maximum of source and destination
    MAX = 0x8008

    # Provoking vertex

    #: Specifies the first vertex should be used as the source of data for flat shaded varyings.
    #: Used with :py:attr:`Context.provoking_vertex`.
    FIRST_VERTEX_CONVENTION = 0x8E4D
    #: Specifies the last vertex should be used as the source of data for flat shaded varyings.
    #: Used with :py:attr:`Context.provoking_vertex`.
    LAST_VERTEX_CONVENTION = 0x8E4E

    __slots__ = ['mglo', '_screen', '_info', '_extensions', 'version_code', 'fbo', '_gc_mode', '_objects', 'extra']

    def __init__(self):
        self.mglo = None  #: Internal representation for debug purposes only.
        self._screen = None
        self._info = None
        self._extensions = None
        self.version_code = None  #: int: The OpenGL version code. Reports ``410`` for OpenGL 4.1
        #: Framebuffer: The active framebuffer.
        #: Set every time :py:meth:`Framebuffer.use()` is called.
        self.fbo = None
        self.extra = None  #: Any - Attribute for storing user defined objects
        self._gc_mode = None
        self._objects: Deque[Any] = deque()
        raise TypeError()

    def __repr__(self) -> str:
        return f"<Context {id(self)} version_code={self.version_code}>"

    def __eq__(self, other: Any):
        return type(self) is type(other) and self.mglo is other.mglo

    def __hash__(self) -> int:
        return id(self)

    def __del__(self):
        if hasattr(self, "_gc_mode") and self._gc_mode == "auto":
            self.release()

    @property
    def gc_mode(self) -> Optional[str]:
        """
        Optional[str]: The garbage collection mode.

        The default mode is ``None`` meaning no automatic
        garbage collection is done. Other modes are ``auto``
        and ``context_gc``. Please see documentation for
        the appropriate configuration.

        Examples::

            # Disable automatic garbage collection.
            # Each objects needs to be explicitly released.
            ctx.gc_mode = None

            # Collect all dead objects in the context and
            # release them by calling Context.gc()
            ctx.gc_mode = "context_gc"
            ctx.gc()  # Deletes the collected objects

            # Enable automatic garbage collection like
            # we normally expect in python.
            ctx.gc_mode = "auto"
        """
        return self._gc_mode

    @gc_mode.setter
    def gc_mode(self, value: Optional[str]) -> None:
        if value not in self._valid_gc_modes:
            raise ValueError("Valid  gc modes:", self._valid_gc_modes)

        self._gc_mode = value

    @property
    def objects(self) -> Deque[Any]:
        """
        Moderngl objects scheduled for deletion.

        These are deleted when calling :py:meth:`Context.gc`.
        """
        return self._objects

    def gc(self) -> int:
        """
        Deletes OpenGL objects.

        This method must be called to garbage collect
        OpenGL resources when ``gc_mode`` is ``"context_gc"```.

        Calling this method with any other ``gc_mode`` configuration
        has no effect and is perfectly safe.

        Returns:
            int: Number of objects deleted
        """
        count = 0
        # Keep iterating until there are no more objects.
        # An object deletion can trigger new objects to be added
        while self._objects:
            # Remove the oldest objects first
            obj = self._objects.popleft()
            obj.release()
            count += 1

        return count

    @property
    def line_width(self) -> float:
        """
        float: Set the default line width.

        .. Warning:: A line width other than 1.0 is not guaranteed to work
                     across different OpenGL implementations. For wide
                     lines you should be using geometry shaders.

        """
        return self.mglo.line_width

    @line_width.setter
    def line_width(self, value: float) -> None:
        self.mglo.line_width = value

    @property
    def point_size(self) -> float:
        """
        float: Set/get the point size.

        Point size changes the pixel size of rendered points. The min and max values
        are limited by ``POINT_SIZE_RANGE``.
        This value usually at least ``(1, 100)``, but this depends on the drivers/vendors.

        If variable point size is needed you can enable ``PROGRAM_POINT_SIZE``
        and write to ``gl_PointSize`` in the vertex or geometry shader.

        .. Note::

            Using a geometry shader to create triangle strips from points is often a safer
            way to render large points since you don't have have any size restrictions.
        """
        return self.mglo.point_size

    @point_size.setter
    def point_size(self, value: float) -> None:
        self.mglo.point_size = value

    @property
    def depth_func(self) -> str:
        """
        str: Set the default depth func.

        Example::

            ctx.depth_func = '<='  # GL_LEQUAL
            ctx.depth_func = '<'   # GL_LESS
            ctx.depth_func = '>='  # GL_GEQUAL
            ctx.depth_func = '>'   # GL_GREATER
            ctx.depth_func = '=='  # GL_EQUAL
            ctx.depth_func = '!='  # GL_NOTEQUAL
            ctx.depth_func = '0'   # GL_NEVER
            ctx.depth_func = '1'   # GL_ALWAYS
        """
        raise NotImplementedError()

    @depth_func.setter
    def depth_func(self, value: str) -> None:
        self.mglo.depth_func = value

    @property
    def blend_func(self) -> Tuple[int, int]:
        """
        tuple: Set the blend func (write only).

        Blend func can be set for rgb and alpha separately if needed.

        Supported blend functions are::

            moderngl.ZERO
            moderngl.ONE
            moderngl.SRC_COLOR
            moderngl.ONE_MINUS_SRC_COLOR
            moderngl.DST_COLOR
            moderngl.ONE_MINUS_DST_COLOR
            moderngl.SRC_ALPHA
            moderngl.ONE_MINUS_SRC_ALPHA
            moderngl.DST_ALPHA
            moderngl.ONE_MINUS_DST_ALPHA

            # Shortcuts
            moderngl.DEFAULT_BLENDING     # (SRC_ALPHA, ONE_MINUS_SRC_ALPHA)
            moderngl.ADDITIVE_BLENDING    # (ONE, ONE)
            moderngl.PREMULTIPLIED_ALPHA  # (SRC_ALPHA, ONE)

        Example::

            # For both rgb and alpha
            ctx.blend_func = moderngl.SRC_ALPHA, moderngl.ONE_MINUS_SRC_ALPHA

            # Separate for rgb and alpha
            ctx.blend_func = (
                moderngl.SRC_ALPHA, moderngl.ONE_MINUS_SRC_ALPHA,
                moderngl.ONE, moderngl.ONE
            )
        """
        raise NotImplementedError()

    @blend_func.setter
    def blend_func(self, value: Tuple[int, int]) -> None:
        self.mglo.blend_func = tuple(value)

    @property
    def blend_equation(self) -> Tuple[int, int]:
        """
        tuple: Set the blend equation (write only).

        Blend equations specify how source and destination colors are combined
        in blending operations. By default ``FUNC_ADD`` is used.

        Blend equation can be set for rgb and alpha separately if needed.

        Supported functions are::

            moderngl.FUNC_ADD               # source + destination
            moderngl.FUNC_SUBTRACT          # source - destination
            moderngl.FUNC_REVERSE_SUBTRACT  # destination - source
            moderngl.MIN                    # Minimum of source and destination
            moderngl.MAX                    # Maximum of source and destination

        Example::

            # For both rgb and alpha channel
            ctx.blend_equation = moderngl.FUNC_ADD

            # Separate for rgb and alpha channel
            ctx.blend_equation = moderngl.FUNC_ADD, moderngl.MAX
        """
        raise NotImplementedError()

    @blend_equation.setter
    def blend_equation(self, value: Tuple[int, int]) -> None:
        if not isinstance(value, tuple):
            self.mglo.blend_equation = tuple([value])
        else:
            self.mglo.blend_equation = tuple(value)

    @property
    def multisample(self) -> bool:
        """
        bool: Enable/disable multisample mode (``GL_MULTISAMPLE``).

        This property is write only.

        Example::

            # Enable
            ctx.multisample = True
            # Disable
            ctx.multisample = False
        """
        raise NotImplementedError()

    @multisample.setter
    def multisample(self, value: bool) -> None:
        self.mglo.multisample = value

    @property
    def provoking_vertex(self) -> int:
        """
        int: Specifies the vertex to be used as the source of data for flat shaded varyings.

        Flatshading a vertex shader varying output (ie. ``flat out vec3 pos``) means to assign
        all vetices of the primitive the same value for that output. The vertex from which
        these values is derived is known as the provoking vertex.

        It can be configured to be the first or the last vertex.

        This property is write only.

        Example::

            # Use first vertex
            ctx.provoking_vertex = moderngl.FIRST_VERTEX_CONVENTION

            # Use last vertex
            ctx.provoking_vertex = moderngl.LAST_VERTEX_CONVENTION
        """
        raise NotImplementedError()

    @provoking_vertex.setter
    def provoking_vertex(self, value: int) -> None:
        self.mglo.provoking_vertex = value

    @property
    def polygon_offset(self) -> Tuple[float, float]:
        """
        tuple: Get or set the current polygon offset.

        The tuple values represents two float values: ``unit`` and a ``factor``::

            ctx.polygon_offset = unit, factor

        When drawing polygons, lines or points directly on top of
        exiting geometry the result is often not visually pleasant.
        We can experience z-fighting or partially fading fragments
        due to different primitives not being rasterized in the exact
        same way or simply depth buffer precision issues.

        For example when visualizing polygons drawing a wireframe
        version on top of the original mesh, these issues are
        immediately apparent. Applying decals to surfaces is
        another common example.

        The official documentation states the following::

            When enabled, the depth value of each fragment is added
            to a calculated offset value. The offset is added before
            the depth test is performed and before the depth value
            is written into the depth buffer. The offset value o is calculated by:
            o = m * factor + r * units
            where m is the maximum depth slope of the polygon and r is the smallest
            value guaranteed to produce a resolvable difference in window coordinate
            depth values. The value r is an implementation-specific constant.

        In simpler terms: We use polygon offset to either add a positive offset to
        the geometry (push it away from you) or a negative offset to geometry
        (pull it towards you).

        * ``units`` is a constant offset to depth and will do the job alone
          if we are working with geometry parallel to the near/far plane.
        * The ``factor`` helps you handle sloped geometry (not parallel to near/far plane).

        In most cases you can get away with ``[-1.0, 1.0]`` for both factor and units,
        but definitely play around with the values. When both values are set to ``0``
        polygon offset is disabled internally.

        To just get started with something you can try::

            # Either push the geomtry away or pull it towards you
            # with support for handling small to medium sloped geometry
            ctx.polygon_offset = 1.0, 1.0
            ctx.polygon_offset = -1.0, -1.0

            # Disable polygon offset
            ctx.polygon_offset = 0, 0
        """
        return self.mglo.polygon_offset

    @polygon_offset.setter
    def polygon_offset(self, value: Tuple[float, float]) -> None:
        factor, units = value
        self.mglo.polygon_offset = (float(factor), float(units))

    @property
    def viewport(self) -> Tuple[int, int, int, int]:
        """
        tuple: Get or set the viewport of the active framebuffer.

        Example::

            >>> ctx.viewport
            (0, 0, 1280, 720)
            >>> ctx.viewport = (0, 0, 640, 360)
            >>> ctx.viewport
            (0, 0, 640, 360)

        If no framebuffer is bound ``(0, 0, 0, 0)`` will be returned.
        """
        return self.mglo.fbo.viewport

    @viewport.setter
    def viewport(self, value: Tuple[int, int, int, int]) -> None:
        x, y, w, h = value
        self.mglo.fbo.viewport = (int(x), int(y), int(w), int(h))

    @property
    def scissor(self) -> Optional[Tuple[int, int, int, int]]:
        """
        tuple: Get or set the scissor box for the active framebuffer.

        When scissor testing is enabled fragments outside
        the defined scissor box will be discarded. This
        applies to rendered geometry or :py:meth:`Context.clear`.

        Setting is value enables scissor testing in the framebuffer.
        Setting the scissor to ``None`` disables scissor testing
        and reverts the scissor box to match the framebuffer size.

        Example::

            # Enable scissor testing
            >>> ctx.scissor = 100, 100, 200, 100
            # Disable scissor testing
            >>> ctx.scissor = None

        If no framebuffer is bound ``(0, 0, 0, 0)`` will be returned.
        """
        return self.mglo.fbo.scissor

    @scissor.setter
    def scissor(self, value: Optional[Tuple[int, int, int, int]]) -> None:
        if value is None:
            self.mglo.fbo.scissor = None
        else:
            x, y, w, h = value
            self.mglo.fbo.scissor = (int(x), int(y), int(w), int(h))

    @property
    def max_samples(self) -> int:
        """int: The maximum supported number of samples for multisampling."""
        return self.mglo.max_samples

    @property
    def max_integer_samples(self) -> int:
        """int: The max integer samples."""
        return self.mglo.max_integer_samples

    @property
    def max_texture_units(self) -> int:
        """int: The max texture units."""
        return self.mglo.max_texture_units

    @property
    def default_texture_unit(self) -> int:
        """int: The default texture unit."""
        return self.mglo.default_texture_unit

    @default_texture_unit.setter
    def default_texture_unit(self, value: int) -> None:
        self.mglo.default_texture_unit = value

    @property
    def max_anisotropy(self) -> float:
        """float: The maximum value supported for anisotropic filtering."""
        return self.mglo.max_anisotropy

    @property
    def screen(self) -> 'Framebuffer':
        """
        Framebuffer: A Framebuffer instance representing the screen.

        Normally set when creating a context with ``create_context()`` attaching to
        an existing context. This is the special system framebuffer
        represented by framebuffer ``id=0``.

        When creating a standalone context this property is not set since
        there are no default framebuffer.
        """
        return self._screen

    @property
    def wireframe(self) -> bool:
        """bool: Wireframe settings for debugging."""
        return self.mglo.wireframe

    @wireframe.setter
    def wireframe(self, value: bool) -> None:
        self.mglo.wireframe = value

    @property
    def front_face(self) -> str:
        """
        str: The front_face. Acceptable values are ``'ccw'`` (default) or ``'cw'``.

        Face culling must be enabled for this to have any effect:
        ``ctx.enable(moderngl.CULL_FACE)``.

        Example::

            # Triangles winded counter-clockwise considered front facing
            ctx.front_face = 'ccw'
            # Triangles winded clockwise considered front facing
            ctx.front_face = 'cw'
        """
        return self.mglo.front_face

    @front_face.setter
    def front_face(self, value: str) -> None:
        self.mglo.front_face = str(value)

    @property
    def cull_face(self) -> str:
        """
        str: The face side to cull. Acceptable values are ``'back'`` (default) ``'front'`` or ``'front_and_back'``.

        This is similar to :py:meth:`Context.front_face`

        Face culling must be enabled for this to have any effect:
        ``ctx.enable(moderngl.CULL_FACE)``.

        Example::

            ctx.cull_face = 'front'
            ctx.cull_face = 'back'
            ctx.cull_face = 'front_and_back'
        """
        return self.mglo.cull_face

    @cull_face.setter
    def cull_face(self, value: str) -> None:
        self.mglo.cull_face = str(value)

    @property
    def patch_vertices(self) -> int:
        """int: The number of vertices that will be used to make up a single patch primitive."""
        return self.mglo.patch_vertices

    @patch_vertices.setter
    def patch_vertices(self, value: int) -> None:
        self.mglo.patch_vertices = value

    @property
    def error(self) -> str:
        """
        str: The result of ``glGetError()`` but human readable.

        This values is provided for debug purposes only and is likely to
        reduce performace when used in a draw loop.
        """
        return self.mglo.error

    @property
    def extensions(self) -> Set[str]:
        """
        Set[str]: The extensions supported by the context.

        All extensions names have a ``GL_`` prefix, so if the spec refers to ``ARB_compute_shader``
        we need to look for ``GL_ARB_compute_shader``::

            # If compute shaders are supported ...
            >> "GL_ARB_compute_shader" in ctx.extensions
            True

        Example data::

            {
                'GL_ARB_multi_bind',
                'GL_ARB_shader_objects',
                'GL_ARB_half_float_vertex',
                'GL_ARB_map_buffer_alignment',
                'GL_ARB_arrays_of_arrays',
                'GL_ARB_pipeline_statistics_query',
                'GL_ARB_provoking_vertex',
                'GL_ARB_gpu_shader5',
                'GL_ARB_uniform_buffer_object',
                'GL_EXT_blend_equation_separate',
                'GL_ARB_tessellation_shader',
                'GL_ARB_multi_draw_indirect',
                'GL_ARB_multisample',
                .. etc ..
            }
        """
        if self._extensions is None:
            self._extensions = self.mglo.extensions

        return self._extensions

    @property
    def info(self) -> Dict[str, Any]:
        """
        dict: OpenGL Limits and information about the context.

        Example::

            # The maximum width and height of a texture
            >> ctx.info["GL_MAX_TEXTURE_SIZE"]
            16384

            # Vendor and renderer
            >> ctx.info["GL_VENDOR"]
            NVIDIA Corporation
            >> ctx.info["GL_RENDERER"]
            NVIDIA GeForce GT 650M OpenGL Engine

        Example data::

            {
                'GL_VENDOR': 'NVIDIA Corporation',
                'GL_RENDERER': 'NVIDIA GeForce GT 650M OpenGL Engine',
                'GL_VERSION': '4.1 NVIDIA-10.32.0 355.11.10.10.40.102',
                'GL_POINT_SIZE_RANGE': (1.0, 2047.0),
                'GL_SMOOTH_LINE_WIDTH_RANGE': (0.5, 1.0),
                'GL_ALIASED_LINE_WIDTH_RANGE': (1.0, 1.0),
                'GL_POINT_FADE_THRESHOLD_SIZE': 1.0,
                'GL_POINT_SIZE_GRANULARITY': 0.125,
                'GL_SMOOTH_LINE_WIDTH_GRANULARITY': 0.125,
                'GL_MIN_PROGRAM_TEXEL_OFFSET': -8.0,
                'GL_MAX_PROGRAM_TEXEL_OFFSET': 7.0,
                'GL_MINOR_VERSION': 1,
                'GL_MAJOR_VERSION': 4,
                'GL_SAMPLE_BUFFERS': 0,
                'GL_SUBPIXEL_BITS': 8,
                'GL_CONTEXT_PROFILE_MASK': 1,
                'GL_UNIFORM_BUFFER_OFFSET_ALIGNMENT': 256,
                'GL_DOUBLEBUFFER': False,
                'GL_STEREO': False,
                'GL_MAX_VIEWPORT_DIMS': (16384, 16384),
                'GL_MAX_3D_TEXTURE_SIZE': 2048,
                'GL_MAX_ARRAY_TEXTURE_LAYERS': 2048,
                'GL_MAX_CLIP_DISTANCES': 8,
                'GL_MAX_COLOR_ATTACHMENTS': 8,
                'GL_MAX_COLOR_TEXTURE_SAMPLES': 8,
                'GL_MAX_COMBINED_FRAGMENT_UNIFORM_COMPONENTS': 233472,
                'GL_MAX_COMBINED_GEOMETRY_UNIFORM_COMPONENTS': 231424,
                'GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS': 80,
                'GL_MAX_COMBINED_UNIFORM_BLOCKS': 70,
                'GL_MAX_COMBINED_VERTEX_UNIFORM_COMPONENTS': 233472,
                'GL_MAX_CUBE_MAP_TEXTURE_SIZE': 16384,
                'GL_MAX_DEPTH_TEXTURE_SAMPLES': 8,
                'GL_MAX_DRAW_BUFFERS': 8,
                'GL_MAX_DUAL_SOURCE_DRAW_BUFFERS': 1,
                'GL_MAX_ELEMENTS_INDICES': 150000,
                'GL_MAX_ELEMENTS_VERTICES': 1048575,
                'GL_MAX_FRAGMENT_INPUT_COMPONENTS': 128,
                'GL_MAX_FRAGMENT_UNIFORM_COMPONENTS': 4096,
                'GL_MAX_FRAGMENT_UNIFORM_VECTORS': 1024,
                'GL_MAX_FRAGMENT_UNIFORM_BLOCKS': 14,
                'GL_MAX_GEOMETRY_INPUT_COMPONENTS': 128,
                'GL_MAX_GEOMETRY_OUTPUT_COMPONENTS': 128,
                'GL_MAX_GEOMETRY_TEXTURE_IMAGE_UNITS': 16,
                'GL_MAX_GEOMETRY_UNIFORM_BLOCKS': 14,
                'GL_MAX_GEOMETRY_UNIFORM_COMPONENTS': 2048,
                'GL_MAX_INTEGER_SAMPLES': 1,
                'GL_MAX_SAMPLES': 8,
                'GL_MAX_RECTANGLE_TEXTURE_SIZE': 16384,
                'GL_MAX_RENDERBUFFER_SIZE': 16384,
                'GL_MAX_SAMPLE_MASK_WORDS': 1,
                'GL_MAX_SERVER_WAIT_TIMEOUT': -1,
                'GL_MAX_TEXTURE_BUFFER_SIZE': 134217728,
                'GL_MAX_TEXTURE_IMAGE_UNITS': 16,
                'GL_MAX_TEXTURE_LOD_BIAS': 15,
                'GL_MAX_TEXTURE_SIZE': 16384,
                'GL_MAX_UNIFORM_BUFFER_BINDINGS': 70,
                'GL_MAX_UNIFORM_BLOCK_SIZE': 65536,
                'GL_MAX_VARYING_COMPONENTS': 0,
                'GL_MAX_VARYING_VECTORS': 31,
                'GL_MAX_VARYING_FLOATS': 0,
                'GL_MAX_VERTEX_ATTRIBS': 16,
                'GL_MAX_VERTEX_TEXTURE_IMAGE_UNITS': 16,
                'GL_MAX_VERTEX_UNIFORM_COMPONENTS': 4096,
                'GL_MAX_VERTEX_UNIFORM_VECTORS': 1024,
                'GL_MAX_VERTEX_OUTPUT_COMPONENTS': 128,
                'GL_MAX_VERTEX_UNIFORM_BLOCKS': 14,
                'GL_MAX_VERTEX_ATTRIB_RELATIVE_OFFSET': 0,
                'GL_MAX_VERTEX_ATTRIB_BINDINGS': 0,
                'GL_VIEWPORT_BOUNDS_RANGE': (-32768, 32768),
                'GL_VIEWPORT_SUBPIXEL_BITS': 0,
                'GL_MAX_VIEWPORTS': 16
            }
        """
        if self._info is None:
            self._info = self.mglo.info

        return self._info

    def clear(
        self,
        red: float = 0.0,
        green: float = 0.0,
        blue: float = 0.0,
        alpha: float = 0.0,
        depth: float = 1.0,
        *,
        viewport: Optional[Union[Tuple[int, int], Tuple[int, int, int, int]]] = None,
        color: Optional[Tuple[float, float, float, float]] = None,
    ) -> None:
        """
        Clear the bound framebuffer.

        If a `viewport` passed in, a scissor test will be used to clear the given viewport.
        This viewport take prescense over the framebuffers :py:attr:`~moderngl.Framebuffer.scissor`.
        Clearing can still be done with scissor if no viewport is passed in.

        This method also respects the
        :py:attr:`~moderngl.Framebuffer.color_mask` and
        :py:attr:`~moderngl.Framebuffer.depth_mask`. It can for example be used to only clear
        the depth or color buffer or specific components in the color buffer.

        If the `viewport` is a 2-tuple it will clear the
        ``(0, 0, width, height)`` where ``(width, height)`` is the 2-tuple.

        If the `viewport` is a 4-tuple it will clear the given viewport.

        Args:
            red (float): color component.
            green (float): color component.
            blue (float): color component.
            alpha (float): alpha component.
            depth (float): depth value.

        Keyword Args:
            viewport (tuple): The viewport.
            color (tuple): Optional rgba color tuple
        """
        if color is not None:
            red, green, blue, alpha, *_ = tuple(color) + (0.0, 0.0, 0.0, 0.0)

        self.mglo.fbo.clear(red, green, blue, alpha, depth, viewport)

    def enable_only(self, flags: int) -> None:
        """
        Clears all existing flags applying new ones.

        Note that the enum values defined in moderngl
        are not the same as the ones in opengl.
        These are defined as bit flags so we can logical
        `or` them together.

        Available flags:

        - :py:data:`moderngl.NOTHING`
        - :py:data:`moderngl.BLEND`
        - :py:data:`moderngl.DEPTH_TEST`
        - :py:data:`moderngl.CULL_FACE`
        - :py:data:`moderngl.RASTERIZER_DISCARD`
        - :py:data:`moderngl.PROGRAM_POINT_SIZE`

        Examples::

            # Disable all flags
            ctx.enable_only(moderngl.NOTHING)

            # Ensure only depth testing and face culling is enabled
            ctx.enable_only(moderngl.DEPTH_TEST | moderngl.CULL_FACE)

        Args:
            flags (EnableFlag): The flags to enable
        """
        self.mglo.enable_only(flags)

    def enable(self, flags: int) -> None:
        """
        Enable flags.

        Note that the enum values defined in moderngl
        are not the same as the ones in opengl.
        These are defined as bit flags so we can logical
        `or` them together.

        For valid flags, please see :py:meth:`enable_only`.

        Examples::

            # Enable a single flag
            ctx.enable(moderngl.DEPTH_TEST)

            # Enable multiple flags
            ctx.enable(moderngl.DEPTH_TEST | moderngl.CULL_FACE | moderngl.BLEND)

        Args:
            flag (int): The flags to enable.
        """
        self.mglo.enable(flags)

    def disable(self, flags: int) -> None:
        """
        Disable flags.

        For valid flags, please see :py:meth:`enable_only`.

        Examples::

            # Only disable depth testing
            ctx.disable(moderngl.DEPTH_TEST)

            # Disable depth testing and face culling
            ctx.disable(moderngl.DEPTH_TEST | moderngl.CULL_FACE)

        Args:
            flag (int): The flags to disable.
        """
        self.mglo.disable(flags)

    def enable_direct(self, enum: int) -> None:
        """
        Gives direct access to ``glEnable`` so unsupported capabilities in ModernGL can be enabled.

        Do not use this to set already supported context flags.

        Example::

            # Enum value from the opengl registry
            GL_CONSERVATIVE_RASTERIZATION_NV = 0x9346
            ctx.enable_direct(GL_CONSERVATIVE_RASTERIZATION_NV)
        """
        self.mglo.enable_direct(enum)

    def disable_direct(self, enum: int) -> None:
        """
        Gives direct access to ``glDisable`` so unsupported capabilities in ModernGL can be disabled.

        Do not use this to set already supported context flags.

        Example::

            # Enum value from the opengl registry
            GL_CONSERVATIVE_RASTERIZATION_NV = 0x9346
            ctx.disable_direct(GL_CONSERVATIVE_RASTERIZATION_NV)
        """
        self.mglo.disable_direct(enum)

    def finish(self) -> None:
        """Wait for all drawing commands to finish."""
        self.mglo.finish()

    def copy_buffer(
        self,
        dst: Buffer,
        src: Buffer,
        size: int = -1,
        *,
        read_offset: int = 0,
        write_offset: int = 0
    ) -> None:
        """
        Copy buffer content.

        Args:
            dst (Buffer): The destination buffer.
            src (Buffer): The source buffer.
            size (int): The number of bytes to copy.

        Keyword Args:
            read_offset (int): The read offset.
            write_offset (int): The write offset.
        """
        self.mglo.copy_buffer(dst.mglo, src.mglo, size, read_offset, write_offset)

    def copy_framebuffer(
        self,
        dst: Union[Framebuffer, Texture],
        src: Framebuffer,
    ) -> None:
        """
        Copy framebuffer content.

        Use this method to:

            - blit framebuffers.
            - copy framebuffer content into a texture.
            - downsample framebuffers. (it will allow to read the framebuffer's content)
            - downsample a framebuffer directly to a texture.

        Args:
            dst (Framebuffer or Texture): Destination framebuffer or texture.
            src (Framebuffer): Source framebuffer.
        """
        self.mglo.copy_framebuffer(dst.mglo, src.mglo)

    def detect_framebuffer(self, glo: Optional[int] = None) -> 'Framebuffer':
        """
        Detect a framebuffer.

        This is already done when creating a context,
        but if the underlying window library for some changes the default framebuffer
        during the lifetime of the application this might be necessary.

        Args:
            glo (int): Framebuffer object.

        Returns:
            :py:class:`Framebuffer` object
        """
        res = Framebuffer.__new__(Framebuffer)
        res.mglo, res._size, res._samples, res._glo = self.mglo.detect_framebuffer(glo)
        res._color_attachments = None
        res._depth_attachment = None
        res.ctx = self
        res._is_reference = True
        res.extra = None
        return res

    def buffer(
        self,
        data: Optional[Any] = None,
        *,
        reserve: int = 0,
        dynamic: bool = False,
    ) -> Buffer:
        """
        Create a :py:class:`Buffer` object.

        Args:
            data (bytes): Content of the new buffer.

        Keyword Args:
            reserve (int): The number of bytes to reserve.
            dynamic (bool): Treat buffer as dynamic.

        Returns:
            :py:class:`Buffer` object
        """
        if type(reserve) is str:
            reserve = mgl.strsize(reserve)

        res = Buffer.__new__(Buffer)
        res.mglo, res._size, res._glo = self.mglo.buffer(data, reserve, dynamic)
        res._dynamic = dynamic
        res.ctx = self
        res.extra = None
        return res

    def external_texture(
        self,
        glo: int,
        size: Tuple[int, int],
        components: int,
        samples: int,
        dtype: str,
    ) -> 'Texture':
        """
        Create a :py:class:`Texture` object from an existing OpenGL texture object.

        Args:
            glo (int): External OpenGL texture object.
            size (tuple): The width and height of the texture.
            components (int): The number of components 1, 2, 3 or 4.
            samples (int): The number of samples. Value 0 means no multisample format.
            dtype (str): Data type.
        """

        res = Texture.__new__(Texture)
        res.mglo, res._glo = self.mglo.external_texture(glo, size, components, samples, dtype)
        res._size = size
        res._components = components
        res._samples = samples
        res._dtype = dtype
        res._depth = False
        res.ctx = self
        res.extra = None
        return res

    def texture(
        self,
        size: Tuple[int, int],
        components: int,
        data: Optional[Any] = None,
        *,
        samples: int = 0,
        alignment: int = 1,
        dtype: str = 'f1',
        internal_format: Optional[int] = None,
    ) -> 'Texture':
        """
        Create a :py:class:`Texture` object.

        .. Warning:: Do not play with ``internal_format`` unless you know exactly
                        you are doing. This is an override to support sRGB and
                        compressed textures if needed.

        Args:
            size (tuple): The width and height of the texture.
            components (int): The number of components 1, 2, 3 or 4.
            data (bytes): Content of the texture.

        Keyword Args:
            samples (int): The number of samples. Value 0 means no multisample format.
            alignment (int): The byte alignment 1, 2, 4 or 8.
            dtype (str): Data type.
            internal_format (int): Override the internalformat of the texture (IF needed)

        Returns:
            :py:class:`Texture` object
        """
        res = Texture.__new__(Texture)
        res.mglo, res._glo = self.mglo.texture(size, components, data, samples, alignment, dtype, internal_format or 0)
        res._size = size
        res._components = components
        res._samples = samples
        res._dtype = dtype
        res._depth = False
        res.ctx = self
        res.extra = None
        return res

    def texture_array(
        self,
        size: Tuple[int, int, int],
        components: int,
        data: Optional[Any] = None,
        *,
        alignment: int = 1,
        dtype: str = 'f1',
    ) -> 'TextureArray':
        """
        Create a :py:class:`TextureArray` object.

        Args:
            size (tuple): The ``(width, height, layers)`` of the texture.
            components (int): The number of components 1, 2, 3 or 4.
            data (bytes): Content of the texture. The size must be ``(width, height * layers)``
                            so each layer is stacked vertically.

        Keyword Args:
            alignment (int): The byte alignment 1, 2, 4 or 8.
            dtype (str): Data type.

        Returns:
            :py:class:`Texture3D` object
        """
        res = TextureArray.__new__(TextureArray)
        res.mglo, res._glo = self.mglo.texture_array(size, components, data, alignment, dtype)
        res._size = size
        res._components = components
        res._dtype = dtype
        res.ctx = self
        res.extra = None
        return res

    def texture3d(
        self,
        size: Tuple[int, int, int],
        components: int,
        data: Optional[Any] = None,
        *,
        alignment: int = 1,
        dtype: str = 'f1',
    ) -> 'Texture3D':
        """
        Create a :py:class:`Texture3D` object.

        Args:
            size (tuple): The width, height and depth of the texture.
            components (int): The number of components 1, 2, 3 or 4.
            data (bytes): Content of the texture.

        Keyword Args:
            alignment (int): The byte alignment 1, 2, 4 or 8.
            dtype (str): Data type.

        Returns:
            :py:class:`Texture3D` object
        """
        res = Texture3D.__new__(Texture3D)
        res.mglo, res._glo = self.mglo.texture3d(size, components, data, alignment, dtype)
        res.ctx = self
        res.extra = None
        return res

    def texture_cube(
        self,
        size: Tuple[int, int],
        components: int,
        data: Optional[Any] = None,
        *,
        alignment: int = 1,
        dtype: str = 'f1',
        internal_format: Optional[int] = None,
    ) -> 'TextureCube':
        """
        Create a :py:class:`TextureCube` object.

        Note that the width and height of the cubemap must be the same
        unless you are using a non-standard extension.

        Args:
            size (tuple): The width, height of the texture. Each side of
                            the cube will have this size.
            components (int): The number of components 1, 2, 3 or 4.
            data (bytes): Content of the texture. The data should be have the following ordering:
                            positive_x, negative_x, positive_y, negative_y, positive_z, negative_z

        Keyword Args:
            alignment (int): The byte alignment 1, 2, 4 or 8.
            dtype (str): Data type.
            internal_format (int): Override the internalformat of the texture (IF needed)

        Returns:
            :py:class:`TextureCube` object
        """
        res = TextureCube.__new__(TextureCube)
        res.mglo, res._glo = self.mglo.texture_cube(size, components, data, alignment, dtype, internal_format or 0)
        res._size = size
        res._components = components
        res._dtype = dtype
        res.ctx = self
        res.extra = None
        return res

    def depth_texture(
        self,
        size: Tuple[int, int],
        data: Optional[Any] = None,
        *,
        samples: int = 0,
        alignment: int = 4,
    ) -> 'Texture':
        """
        Create a :py:class:`Texture` object.

        Args:
            size (tuple): The width and height of the texture.
            data (bytes): Content of the texture.

        Keyword Args:
            samples (int): The number of samples. Value 0 means no multisample format.
            alignment (int): The byte alignment 1, 2, 4 or 8.

        Returns:
            :py:class:`Texture` object
        """
        res = Texture.__new__(Texture)
        res.mglo, res._glo = self.mglo.depth_texture(size, data, samples, alignment)
        res._size = size
        res._components = 1
        res._samples = samples
        res._dtype = 'f4'
        res._depth = True
        res.ctx = self
        res.extra = None
        return res

    def vertex_array(self, *args, **kwargs) -> 'VertexArray':
        """
        Create a :py:class:`VertexArray` object.

        The vertex array describes how buffers are read by a shader program.
        We need to supply buffer formats and attributes names. The attribute names
        are defined by the user in the glsl code and can be anything.

        Examples::

            # Empty vertext array (no attribute input)
            vao = ctx.vertex_array(program)

            # Simple version with a single buffer
            vao = ctx.vertex_array(program, buffer, "in_position", "in_normal")
            vao = ctx.vertex_array(program, buffer, "in_position", "in_normal", index_buffer=ibo)

            # Multiple buffers
            vao = ctx.vertex_array(program, [
                (buffer1, '3f', 'in_position'),
                (buffer2, '3f', 'in_normal'),
            ])
            vao = ctx.vertex_array(program, [
                    (buffer1, '3f', 'in_position'),
                    (buffer2, '3f', 'in_normal'),
                ],
                index_buffer=ibo,
                index_element_size=2,  # 16 bit / 'u2' index buffer
            )

        This method also supports arguments for :py:meth:`Context.simple_vertex_array`.

        Args:
            program (Program): The program used when rendering
            content (list): A list of (buffer, format, attributes).
                            See :ref:`buffer-format-label`.

        Keyword Args:
            index_buffer (Buffer): An index buffer (optional)
            index_element_size (int): byte size of each index element, 1, 2 or 4.
            skip_errors (bool): Ignore errors during creation
            mode (int): The default draw mode (for example: ``TRIANGLES``)

        Returns:
            :py:class:`VertexArray` object
        """
        if len(args) > 2 and type(args[1]) is Buffer:
            return self.simple_vertex_array(*args, **kwargs)
        return self._vertex_array(*args, **kwargs)

    def _vertex_array(
        self,
        program: Program,
        content: Any,
        index_buffer: Optional[Buffer] = None,
        index_element_size: int = 4,
        *,
        skip_errors: bool = False,
        mode: Optional[int] = None,
    ) -> 'VertexArray':
        """
        Create a :py:class:`VertexArray` object.

        Args:
            program (Program): The program used when rendering.
            content (list): A list of (buffer, format, attributes).
                            See :ref:`buffer-format-label`.
            index_buffer (Buffer): An index buffer.

        Keyword Args:
            index_element_size (int): byte size of each index element, 1, 2 or 4.
            skip_errors (bool): Ignore skip_errors varyings.
            mode (int): The default draw mode (for example: ``TRIANGLES``)

        Returns:
            :py:class:`VertexArray` object
        """
        members = program._members
        index_buffer_mglo = None if index_buffer is None else index_buffer.mglo
        mgl_content = tuple(
            (a.mglo, b) + tuple(members.get(x) for x in c)
            for a, b, *c in content
        )

        res = VertexArray.__new__(VertexArray)
        res.mglo, res._glo = self.mglo.vertex_array(
            program.mglo, mgl_content, index_buffer_mglo,
            index_element_size, skip_errors,
        )
        res._program = program
        res._index_buffer = index_buffer
        res._content = content
        res._index_element_size = index_element_size
        if mode is not None:
            res._mode = mode
        else:
            res._mode = self.POINTS if program.is_transform else self.TRIANGLES
        res.ctx = self
        res.extra = None
        res.scope = None
        return res

    def simple_vertex_array(
        self,
        program: Program,
        buffer: Buffer,
        *attributes: Union[List[str], Tuple[str, ...]],
        index_buffer: Optional[Buffer] = None,
        index_element_size: int = 4,
        mode: Optional[int] = None,
    ) -> 'VertexArray':
        """
        Create a :py:class:`VertexArray` object.

        .. Warning:: This method is deprecated and may be removed in the future.
                        Use :py:meth:`Context.vertex_array` instead. It also supports
                        the argument format this method describes.

        Args:
            program (Program): The program used when rendering.
            buffer (Buffer): The buffer.
            attributes (list): A list of attribute names.

        Keyword Args:
            index_element_size (int): byte size of each index element, 1, 2 or 4.
            index_buffer (Buffer): An index buffer.
            mode (int): The default draw mode (for example: ``TRIANGLES``)

        Returns:
            :py:class:`VertexArray` object
        """
        if type(buffer) is list:
            raise SyntaxError('Change simple_vertex_array to vertex_array')

        content = [(buffer, detect_format(program, attributes)) + attributes]
        return self._vertex_array(program, content, index_buffer, index_element_size, mode=mode)

    def program(
        self,
        *,
        vertex_shader: str,
        fragment_shader: Optional[str] = None,
        geometry_shader: Optional[str] = None,
        tess_control_shader: Optional[str] = None,
        tess_evaluation_shader: Optional[str] = None,
        varyings: Tuple[str, ...] = (),
        fragment_outputs: Optional[Dict[str, int]] = None,
        varyings_capture_mode: str = 'interleaved',
    ) -> 'Program':
        """
        Create a :py:class:`Program` object.

        The ``varyings`` are only used when a transform program is created
        to specify the names of the output varyings to capture in the output buffer.

        ``fragment_outputs`` can be used to programmatically map named fragment
        shader outputs to a framebuffer attachment numbers. This can also be done
        by using ``layout(location=N)`` in the fragment shader.

        Args:
            vertex_shader (str): The vertex shader source.
            fragment_shader (str): The fragment shader source.
            geometry_shader (str): The geometry shader source.
            tess_control_shader (str): The tessellation control shader source.
            tess_evaluation_shader (str): The tessellation evaluation shader source.
            varyings (list): A list of varyings.
            fragment_outputs (dict): A dictionary of fragment outputs.
        Returns:
            :py:class:`Program` object
        """

        if varyings_capture_mode not in ('interleaved', 'separate'):
            raise ValueError('varyings_capture_mode must be interleaved or separate')

        if type(varyings) is str:
            varyings = (varyings,)  # type: ignore

        varyings = tuple(varyings)

        if fragment_outputs is None:
            fragment_outputs = {}

        res = Program.__new__(Program)
        res.mglo, res._members, res._subroutines, res._geom, res._glo = self.mglo.program(
            vertex_shader, fragment_shader, geometry_shader, tess_control_shader, tess_evaluation_shader,
            varyings, fragment_outputs, varyings_capture_mode == 'interleaved'
        )

        res._is_transform = fragment_shader is None
        res.ctx = self
        res.extra = None
        return res

    def query(
        self,
        *,
        samples: bool = False,
        any_samples: bool = False,
        time: bool = False,
        primitives: bool = False,
    ) -> 'Query':
        """
        Create a :py:class:`Query` object.

        Keyword Args:
            samples (bool): Query ``GL_SAMPLES_PASSED`` or not.
            any_samples (bool): Query ``GL_ANY_SAMPLES_PASSED`` or not.
            time (bool): Query ``GL_TIME_ELAPSED`` or not.
            primitives (bool): Query ``GL_PRIMITIVES_GENERATED`` or not.
        """
        res = Query.__new__(Query)
        res.mglo = self.mglo.query(samples, any_samples, time, primitives)
        res.crender = None

        if samples or any_samples:
            res.crender = ConditionalRender.__new__(ConditionalRender)
            res.crender.mglo = res.mglo

        res.ctx = self
        res.extra = None
        return res

    def scope(
        self,
        framebuffer: Optional[Framebuffer] = None,
        enable_only: Optional[int] = None,
        *,
        textures: Tuple[Tuple[Texture, int], ...] = (),
        uniform_buffers: Tuple[Tuple[Buffer, int], ...] = (),
        storage_buffers: Tuple[Tuple[Buffer, int], ...] = (),
        samplers: Tuple[Tuple[Sampler, int], ...] = (),
        enable: Optional[int] = None,
    ) -> 'Scope':
        """
        Create a :py:class:`Scope` object.

        Args:
            framebuffer (Framebuffer): The framebuffer to use when entering.
            enable_only (int): The enable_only flags to set when entering.

        Keyword Args:
            textures (tuple): List of (texture, binding) tuples.
            uniform_buffers (tuple): Tuple of (buffer, binding) tuples.
            storage_buffers (tuple): Tuple of (buffer, binding) tuples.
            samplers (tuple): Tuple of sampler bindings
            enable (int): Flags to enable for this vao such as depth testing and blending
        """
        if enable is not None:
            enable_only = enable

        if framebuffer is None:
            framebuffer = self.screen
            if framebuffer is None:
                raise RuntimeError('A framebuffer must be specified')

        mgl_textures = tuple((tex.mglo, idx) for tex, idx in textures)
        mgl_uniform_buffers = tuple((buf.mglo, idx) for buf, idx in uniform_buffers)
        mgl_storage_buffers = tuple((buf.mglo, idx) for buf, idx in storage_buffers)

        res = Scope.__new__(Scope)
        res.mglo = self.mglo.scope(framebuffer.mglo, enable_only, mgl_textures,
                                   mgl_uniform_buffers, mgl_storage_buffers, samplers)
        res.ctx = self
        res._framebuffer = framebuffer
        res._textures = textures
        res._uniform_buffers = uniform_buffers
        res._storage_buffers = storage_buffers
        res._samplers = samplers
        res.extra = None
        return res

    def simple_framebuffer(
        self,
        size: Tuple[int, int],
        components: int = 4,
        *,
        samples: int = 0,
        dtype: str = 'f1',
    ) -> 'Framebuffer':
        """
        Creates a :py:class:`Framebuffer` with a single color attachment \
        and depth buffer using :py:class:`moderngl.Renderbuffer` attachments.

        Args:
            size (tuple): The width and height of the renderbuffer.
            components (int): The number of components 1, 2, 3 or 4.

        Keyword Args:
            samples (int): The number of samples. Value 0 means no multisample format.
            dtype (str): Data type.

        Returns:
            :py:class:`Framebuffer` object
        """
        return self.framebuffer(
            self.renderbuffer(size, components, samples=samples, dtype=dtype),
            self.depth_renderbuffer(size, samples=samples),
        )

    def framebuffer(
        self,
        color_attachments: Any = (),
        depth_attachment: Optional[Union[Texture, Renderbuffer]] = None,
    ) -> 'Framebuffer':
        """
        A :py:class:`Framebuffer` is a collection of buffers that can be \
        used as the destination for rendering. The buffers for Framebuffer \
        objects reference images from either Textures or Renderbuffers.

        Args:
            color_attachments (list): A list of :py:class:`Texture` or
                                        :py:class:`Renderbuffer` objects.
            depth_attachment (Renderbuffer or Texture): The depth attachment.

        Returns:
            :py:class:`Framebuffer` object
        """
        if type(color_attachments) is Texture or type(color_attachments) is Renderbuffer:
            color_attachments = (color_attachments,)

        ca_mglo = tuple(x.mglo for x in color_attachments)
        da_mglo = None if depth_attachment is None else depth_attachment.mglo

        res = Framebuffer.__new__(Framebuffer)
        res.mglo, res._size, res._samples, res._glo = self.mglo.framebuffer(ca_mglo, da_mglo)
        res._color_attachments = tuple(color_attachments)
        res._depth_attachment = depth_attachment
        res.ctx = self
        res._is_reference = False
        res.extra = None
        return res

    def renderbuffer(
        self,
        size: Tuple[int, int],
        components: int = 4,
        *,
        samples: int = 0,
        dtype: str = 'f1',
    ) -> 'Renderbuffer':
        """
        :py:class:`Renderbuffer` objects are OpenGL objects that contain images. \
        They are created and used specifically with :py:class:`Framebuffer` objects.

        Args:
            size (tuple): The width and height of the renderbuffer.
            components (int): The number of components 1, 2, 3 or 4.

        Keyword Args:
            samples (int): The number of samples. Value 0 means no multisample format.
            dtype (str): Data type.

        Returns:
            :py:class:`Renderbuffer` object
        """
        res = Renderbuffer.__new__(Renderbuffer)
        res.mglo, res._glo = self.mglo.renderbuffer(size, components, samples, dtype)
        res._size = size
        res._components = components
        res._samples = samples
        res._dtype = dtype
        res._depth = False
        res.ctx = self
        res.extra = None
        return res

    def depth_renderbuffer(
        self,
        size: Tuple[int, int],
        *,
        samples: int = 0
    ) -> 'Renderbuffer':
        """
        :py:class:`Renderbuffer` objects are OpenGL objects that contain images. \
        They are created and used specifically with :py:class:`Framebuffer` objects.

        Args:
            size (tuple): The width and height of the renderbuffer.

        Keyword Args:
            samples (int): The number of samples. Value 0 means no multisample format.

        Returns:
            :py:class:`Renderbuffer` object
        """
        res = Renderbuffer.__new__(Renderbuffer)
        res.mglo, res._glo = self.mglo.depth_renderbuffer(size, samples)
        res._size = size
        res._components = 1
        res._samples = samples
        res._dtype = 'f4'
        res._depth = True
        res.ctx = self
        res.extra = None
        return res

    def compute_shader(self, source: str) -> 'ComputeShader':
        """
        A :py:class:`ComputeShader` is a Shader Stage that is used entirely \
        for computing arbitrary information. While it can do rendering, it \
        is generally used for tasks not directly related to drawing.

        Args:
            source (str): The source of the compute shader.

        Returns:
            :py:class:`ComputeShader` object
        """
        res = ComputeShader.__new__(ComputeShader)
        res.mglo, res._members, res._glo = self.mglo.compute_shader(source)

        res.ctx = self
        res.extra = None
        return res

    def sampler(
        self,
        repeat_x: bool = True,
        repeat_y: bool = True,
        repeat_z: bool = True,
        filter: Optional[Tuple[int, int]] = None,
        anisotropy: float = 1.0,
        compare_func: str = '?',
        border_color: Optional[Tuple[float, float, float, float]] = None,
        min_lod: float = -1000.0,
        max_lod: float = 1000.0,
        texture: Optional[Texture] = None,
    ) -> Sampler:
        """
        Create a :py:class:`Sampler` object.

        Keyword Arguments:
            repeat_x (bool): Repeat texture on x
            repeat_y (bool): Repeat texture on y
            repeat_z (bool): Repeat texture on z
            filter (tuple): The min and max filter
            anisotropy (float): Number of samples for anisotropic filtering.
                                Any value greater than 1.0 counts as a use of
                                anisotropic filtering
            compare_func: Compare function for depth textures
            border_color (tuple): The (r, g, b, a) color for the texture border.
                                    When this value is set the ``repeat_`` values are
                                    overridden setting the texture wrap to return the
                                    border color when outside ``[0, 1]`` range.
            min_lod (float): Minimum level-of-detail parameter (Default ``-1000.0``).
                                This floating-point value limits the selection of highest
                                resolution mipmap (lowest mipmap level)
            max_lod (float): Minimum level-of-detail parameter (Default ``1000.0``).
                                This floating-point value limits the selection of the
                                lowest resolution mipmap (highest mipmap level)
            texture (Texture): The texture for this sampler
        """
        res = Sampler.__new__(Sampler)
        res.mglo, res._glo = self.mglo.sampler()
        res.ctx = self
        res.repeat_x = repeat_x
        res.repeat_y = repeat_y
        res.repeat_z = repeat_z
        res.filter = filter or (9729, 9729)
        res.anisotropy = anisotropy
        res.compare_func = compare_func
        if border_color:
            res.border_color = border_color
        res.min_lod = min_lod
        res.max_lod = max_lod
        res.extra = None
        res.texture = texture
        return res

    def clear_samplers(self, start: int = 0, end: int = -1) -> None:
        """
        Unbinds samplers from texture units.

        Sampler bindings do clear automatically between every frame,
        but lingering samplers can still be a source of weird bugs during
        the frame rendering. This methods provides a fairly brute force
        and efficient way to ensure texture units are clear.

        Keyword Arguments:

            start (int): The texture unit index to start the clearing samplers
            stop (int): The texture unit index to stop clearing samplers

        Example::

            # Clear texture unit 0, 1, 2, 3, 4
            ctx.clear_samplers(start=0, end=5)

            # Clear texture unit 4, 5, 6, 7
            ctx.clear_samplers(start=4, end=8)
        """
        self.mglo.clear_samplers(start, end)

    def core_profile_check(self) -> None:
        """
        Core profile check.

        FOR DEBUG PURPOSES ONLY
        """
        profile_mask = self.info['GL_CONTEXT_PROFILE_MASK']
        if profile_mask != 1:
            warnings.warn('The window should request a CORE OpenGL profile')

        version_code = self.version_code
        if not version_code:
            major, minor = map(int, self.info['GL_VERSION'].split('.', 2)[:2])
            version_code = major * 100 + minor * 10

        if version_code < 330:
            warnings.warn('The window should support OpenGL 3.3+ (version_code=%d)' % version_code)

    def __enter__(self):
        """
        Enters the context.

        This should ideally be used with the ``with`` statement::

            with other_context as ctx:
                # Do something in this context

        When exiting the context the previously bound context is activated again.

        .. Warning:: Context switching can be risky unless you know what you are doing.
                     Use with care.
        """
        self.mglo.__enter__()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        Exit the context.

        See :py:meth:`Context.__enter__`
        """
        self.mglo.__exit__(exc_type, exc_val, exc_tb)

    def release(self) -> None:
        """
        Release the ModernGL context.

        If the context is not standalone the standard
        backends in ``glcontext`` will not do anything
        because the context was not created by moderngl.

        Standalone contexts can normally be released.
        """
        if self.mglo is not None:
            self.mglo.release()
            self.mglo = InvalidObject()


def create_context(
    require: Optional[int] = None,
    standalone: bool = False,
    share: bool = False,
    **settings: Dict[str, Any],
) -> Context:
    """
    Create a ModernGL context by loading OpenGL functions from an existing OpenGL context. \
    An OpenGL context must exists.

    Example::

        # Accept the current context version
        ctx = moderngl.create_context()

        # Require at least OpenGL 4.3
        ctx = moderngl.create_context(require=430)

        # Create a headless context requiring OpenGL 4.3
        ctx = moderngl.create_context(require=430, standalone=True)

    Keyword Arguments:
        require (int): OpenGL version code (default: 330)
        standalone (bool): Headless flag
        share (bool): Attempt to create a shared context
        **settings: Other backend specific settings

    Returns:
        :py:class:`Context` object
    """
    if require is None:
        require = 330

    mode = 'standalone' if standalone is True else 'detect'
    if share is True:
        mode = 'share'

    from moderngl import mgl  # type: ignore

    ctx = Context.__new__(Context)
    ctx.mglo, ctx.version_code = mgl.create_context(glversion=require, mode=mode, **settings)
    ctx._info = None
    ctx._extensions = None
    ctx.extra = None
    ctx._gc_mode = None
    ctx._objects = deque()

    if ctx.version_code < require:
        raise ValueError('Requested OpenGL version {0}, got version {1}'.format(
            require, ctx.version_code))

    if standalone:
        ctx._screen = None
        ctx.fbo = None
    else:
        ctx._screen = ctx.detect_framebuffer(0)  # Default framebuffer
        ctx.fbo = ctx.detect_framebuffer()  # Currently bound framebuffer
        ctx.mglo.fbo = ctx.fbo.mglo

    return ctx


def create_standalone_context(
    require: Optional[int] = None,
    share: bool = False,
    **settings: Dict[str, Any],
) -> 'Context':
    """
    Create a standalone/headless ModernGL context.

    The preferred way of making a context is through :py:func:`moderngl.create_context`.

    Example::

        # Create a context with highest possible supported version
        ctx = moderngl.create_context()

        # Require at least OpenGL 4.3
        ctx = moderngl.create_context(require=430)

    Keyword Arguments:
        require (int): OpenGL version code.
        share (bool): Attempt to create a shared context
        settings: keyword config values for the context backend

    Returns:
        :py:class:`Context` object
    """
    if require is None:
        require = 330

    mode = 'share' if share is True else 'standalone'

    ctx = Context.__new__(Context)
    ctx.mglo, ctx.version_code = mgl.create_context(glversion=require, mode=mode, **settings)
    ctx._screen = None
    ctx.fbo = None
    ctx._info = None
    ctx._extensions = None
    ctx.extra = None
    ctx._gc_mode = None
    ctx._objects = deque()

    if require is not None and ctx.version_code < require:
        raise ValueError('Requested OpenGL version {0}, got version {1}'.format(
            require, ctx.version_code))

    return ctx



from typing import Optional, Tuple, Final

import moderngl
import numpy as np

from ported._example import Example

NB_PARTICLES: Final[int] = 2 << 10  # 1024
POINT_SIZE: Final[int] = 2 << 3  # 16
# https://www.toutes-les-couleurs.com/code-couleur-rvb.php
CHAMPAGNE_COLOR: Final[Tuple[float, float, float, float]] = (251 / 255., 242 / 255., 183 / 255., 1.0)


class Particles(Example):
    title = "Particle System - Champagne Bubbles"
    gl_version = (3, 3)

    def __init__(
            self,
            shader_transform: Optional[moderngl.Program] = None,
            nb_max_particles: int = NB_PARTICLES,
            **kwargs
    ):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                 #version 330
                 
                 in vec2 in_vert;
                 in float in_hot;

                 out float out_hot;
                 
                 void main() {
                     out_hot = in_hot;
                     gl_Position = vec4(in_vert.xy, 0.0, 1.0);
                 }
             ''',
            fragment_shader='''
                 #version 330
                 
                 uniform vec4 u_color_particle;

                 in float out_hot;

                 out vec4 f_color;

                 void main() {
                     float d_to_center = distance(gl_PointCoord.xy, vec2(0.5, 0.5));
                     if (d_to_center > out_hot*0.5)
                        discard;
                     f_color = vec4(1.0 - vec3(d_to_center), 1.0);

                     f_color *= u_color_particle;
                 }
             ''',
        )

        if shader_transform is None:
            self.transform = self.ctx.program(
                vertex_shader='''
                   #version 330
                   
                   #define M_PI 3.14159265358979323846264
                   // Golden Ratio
                   #define PHI  1.61803398874989484820459 * 00000.1
                   #define PI   M_PI * 00000.1
                   #define SQ2  1.41421356237309504880169 * 10000.0
                   // Gravity factor
                   #define G    0.980
                   // Rate of Decline (particle life)
                   #define ROD  0.006
                   
                   uniform vec2 u_acc;
                   uniform float u_time;

                   in vec2 in_pos;
                   in float in_hot;
                   in vec2 in_vel;
                   in float in_id;

                   out vec2 out_pos;
                   out float out_hot;
                   out vec2 out_vel;
                   out float out_id;
                   
                   // https://www.shadertoy.com/view/ltB3zD
                   float gold_noise(in vec2 coordinate, in float seed){
                       return fract(sin(dot(coordinate*(seed+PHI), vec2(PHI, PI)))*SQ2);
                   }

                   float map(float value, float inMin, float inMax, float outMin, float outMax) {
                     return outMin + (outMax - outMin) * (value - inMin) / (inMax - inMin);
                   }

                   void respawn() {
                       out_hot = map(gold_noise(vec2(in_id, u_time), 25), 0.0, 1.0, 0.4, 1.0);
                       out_pos = vec2(
                           map(gold_noise(vec2(in_id, u_time), 0), 0, 1, -1, +1), 
                           map(gold_noise(vec2(in_id, u_time), 50), 0, 1, -1, +1)
                       );
                       float r = gold_noise(vec2(in_id, u_time), 100) * 0.015;
                       out_vel = vec2(0.0, r);
                   }

                   void update() {
                       out_hot = in_hot - ROD;
                       out_pos = in_pos + in_vel + u_acc;
                       out_vel = in_vel * G;
                   }

                   void main() {
                       // particle is "dead" (out of life) -> respawn
                       if(in_hot <= 0.0) respawn();
                       // particle is out of screen -> respawn
                       else if(abs(in_pos.x) > 1.0) respawn();
                       else if(abs(in_pos.y) > 1.0) respawn();
                       // otherwise -> update particle: position, velocity, live, ...
                       else update();

                       out_id = in_id;
                   }
                ''',
                varyings=['out_pos', 'out_hot', 'out_vel', 'out_id']
            )
        else:
            self.transform = shader_transform

        self.color_particle = self.prog['u_color_particle']
        # https://www.toutes-les-couleurs.com/code-couleur-rvb.php
        self.color_particle.value = CHAMPAGNE_COLOR

        self.nb_particles = nb_max_particles

        self.acc = self.transform['u_acc']
        self.acc.value = (0.0, -0.0001)

        # at init => particles at center with random velocities
        self.vbo1 = self.ctx.buffer(
            b''.join(
                Particles.gen_particle(range_angle_axis=np.pi * 2.0, id_particle=i)
                for i in range(self.nb_particles)
            )
        )
        # vertex buffer/array object use for updating particles system
        self.vbo2 = self.ctx.buffer(reserve=self.vbo1.size)

        self.vao1 = self.ctx.simple_vertex_array(
            self.transform, self.vbo1,
            'in_pos', 'in_hot', 'in_vel', 'in_id'
        )

        self.render_vao = self.ctx.vertex_array(self.prog, [
            # - '2f' => 2 floats ~ vec2: in_vert
            # - '1f' => 1 float: in_hot
            # - '3x4' => drop 3 * 4bytes = 3 floats ~ vec2 + float: in_vel, in_id
            (self.vbo1, '2f 1f 3x4', 'in_vert', 'in_hot'),
        ], skip_errors=False)

        self.u_time = self.transform['u_time']
        self.u_time.value = 0.0

        self.ctx.point_size = float(POINT_SIZE)

    @staticmethod
    def gen_particle(
            position: Tuple[float, float] = (0.0, 0.0),
            angle_main_axis: float = np.pi / 2.0,
            range_angle_axis: float = np.pi / 8.0,
            id_particle: int = 0,
    ) -> bytes:
        """
        Generate particle with random velocity (describe by a random (non unit) vector)
        """
        # random velocity (orientation and velocity)
        a = np.random.uniform(angle_main_axis - range_angle_axis, angle_main_axis + range_angle_axis)
        r = np.random.uniform(0.0, 0.02)
        velocity_x = np.cos(a) * r
        velocity_y = np.sin(a) * r
        # particle at center with a random velocity
        return np.array(
            [
                *position,  # position[vec2]
                1.0,  # hotness[float]
                velocity_x, velocity_y,  # velocity[vec2]
                float(id_particle)  # id particle[float]
            ]
        ).astype('f4').tobytes()  # encoded float32

    def render_particles(self) -> None:
        """
        Render particles system.
        Using a static vertex array object and rendering point sprites.
        """
        self.render_vao.render(moderngl.POINTS, self.nb_particles)

    def update_particles(self, time) -> None:
        """
        Update particles system on GPU
        """
        self.u_time.value = time
        self.vao1.transform(self.vbo2, moderngl.POINTS, self.nb_particles)
        self.ctx.copy_buffer(self.vbo1, self.vbo2)

    def render(self, time, _frame_time) -> None:
        # Physic pass: update particles system
        self.update_particles(time)
        # Rendering pass: render particles system
        self.ctx.clear(*CHAMPAGNE_COLOR)
        self.render_particles()


if __name__ == '__main__':
    Particles.run()



import io

import numpy as np
from PIL import Image

from basic_colors_and_texture import ColorsAndTexture

import matplotlib
matplotlib.use('svg')
import matplotlib.pyplot as plt


class MatplotlibTexture(ColorsAndTexture):
    title = "Matplotlib as Texture"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        figure_size = (640, 360)

        temp = io.BytesIO()
        plt.figure(0, figsize=(figure_size[0] / 72, figure_size[1] / 72))

        mu, sigma = 100, 15
        x = mu + sigma * np.random.randn(10000)
        n, bins, patches = plt.hist(x, 50, density=True, facecolor='r', alpha=0.75)

        plt.axis([40, 160, 0, 0.03])
        plt.grid(True)
        plt.show()

        plt.savefig(temp, format='raw', dpi=72)
        temp.seek(0)

        img = Image.frombytes('RGBA', figure_size, temp.read()).transpose(Image.FLIP_TOP_BOTTOM).convert('RGB')
        self.texture = self.ctx.texture(img.size, 3, img.tobytes())
        self.texture.build_mipmaps()


if __name__ == '__main__':
    MatplotlibTexture.run()



import numpy as np
from pyrr import Matrix44

import moderngl
from ported._example import Example


class InstancedCrates(Example):
    '''
        This example renders 32x32 crates.
        For each crate the location is [x, y, sin(a * time + b)]
        There are 1024 crates aligned in a grid.
    '''
    title = "Instanced Crates"
    gl_version = (3, 3)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                uniform mat4 Mvp;

                in vec3 in_move;

                in vec3 in_position;
                in vec3 in_normal;
                in vec2 in_texcoord_0;

                out vec3 v_vert;
                out vec3 v_norm;
                out vec2 v_text;

                void main() {
                    gl_Position = Mvp * vec4(in_position + in_move, 1.0);
                    v_vert = in_position + in_move;
                    v_norm = in_normal;
                    v_text = in_texcoord_0;
                }
            ''',
            fragment_shader='''
                #version 330

                uniform vec3 Light;
                uniform sampler2D Texture;

                in vec3 v_vert;
                in vec3 v_norm;
                in vec2 v_text;

                out vec4 f_color;

                void main() {
                    float lum = clamp(dot(normalize(Light - v_vert), normalize(v_norm)), 0.0, 1.0) * 0.8 + 0.2;
                    f_color = vec4(texture(Texture, v_text).rgb * lum, 1.0);
                }
            ''',
        )

        self.mvp = self.prog['Mvp']
        self.light = self.prog['Light']

        self.scene = self.load_scene('crate.obj')
        self.texture = self.load_texture_2d('crate.png')

        # Add a new buffer into the VAO wrapper in the scene.
        # This is simply a collection of named buffers that is auto mapped
        # to attributes in the vertex shader with the same name.
        self.instance_data = self.ctx.buffer(reserve=12 * 1024)
        vao_wrapper = self.scene.root_nodes[0].mesh.vao
        vao_wrapper.buffer(self.instance_data, '3f/i', 'in_move')
        # Create the actual vao instance (auto mapping in action)
        self.vao = vao_wrapper.instance(self.prog)

        self.crate_a = np.random.uniform(0.7, 0.8, 32 * 32)
        self.crate_b = np.random.uniform(0.0, 6.3, 32 * 32)
        self.crate_x = (np.tile(np.arange(32), 32) - 16) * 1.5
        self.crate_y = (np.repeat(np.arange(32), 32) - 16) * 1.5
        self.crate_x += np.random.uniform(-0.2, 0.2, 32 * 32)
        self.crate_y += np.random.uniform(-0.2, 0.2, 32 * 32)

    def render(self, time, frame_time):
        angle = time * 0.2
        self.ctx.clear(1.0, 1.0, 1.0)
        self.ctx.enable(moderngl.DEPTH_TEST)

        camera_pos = (np.cos(angle) * 5.0, np.sin(angle) * 5.0, 2.0)

        proj = Matrix44.perspective_projection(45.0, self.aspect_ratio, 0.1, 1000.0)
        lookat = Matrix44.look_at(
            camera_pos,
            (0.0, 0.0, 0.5),
            (0.0, 0.0, 1.0),
        )

        self.mvp.write((proj * lookat).astype('f4'))
        self.light.value = camera_pos

        crate_z = np.sin(self.crate_a * time + self.crate_b) * 0.2
        coordinates = np.dstack([self.crate_x, self.crate_y, crate_z])

        self.instance_data.write(coordinates.astype('f4'))
        self.texture.use()
        self.vao.render(instances=1024)


if __name__ == '__main__':
    InstancedCrates.run()



# """
# NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
# """

# import struct

# import ModernGL
# from PIL import Image

# ctx = ModernGL.create_standalone_context()

# prog = ctx.program(
#     ctx.vertex_shader('''
#         #version 330
#         in vec2 vert;
#         void main() {
#             gl_Position = vec4(vert, 0.0, 1.0);
#         }
#     '''),
#     ctx.fragment_shader('''
#         #version 330
#         out vec4 color;
#         void main() {
#             color = vec4(0.3, 0.5, 1.0, 1.0);
#         }
#     '''),
# ])

# vbo = ctx.buffer(struct.pack('6f', 0.0, 0.8, -0.6, -0.8, 0.6, -0.8))
# vao = ctx.simple_vertex_array(prog, vbo, ['vert'])

# fbo = ctx.framebuffer(ctx.renderbuffer((512, 512)))

# fbo.use()
# fbo.clear(0.9, 0.9, 0.9)
# vao.render()

# pixels = fbo.read(components=3, alignment=1)
# img = Image.frombytes('RGB', fbo.size, pixels).transpose(Image.FLIP_TOP_BOTTOM)
# img.show()



import struct

import ModernGL
from PyQt5 import QtOpenGL, QtWidgets


class QGLControllerWidget(QtOpenGL.QGLWidget):
    def __init__(self):
        fmt = QtOpenGL.QGLFormat()
        fmt.setVersion(3, 3)
        fmt.setProfile(QtOpenGL.QGLFormat.CoreProfile)
        fmt.setSampleBuffers(True)
        super(QGLControllerWidget, self).__init__(fmt, None)

    def initializeGL(self):
        self.ctx = ModernGL.create_context()

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                uniform vec2 WindowSize;

                in vec2 in_vert;
                in vec3 in_color;

                out vec3 v_color;

                void main() {
                    v_color = in_color;
                    gl_Position = vec4(in_vert / WindowSize * 2.0, 0.0, 1.0);
                }
            '''),
            fragment_shader='''
                #version 330

                in vec3 v_color;
                out vec4 f_color;

                void main() {
                    f_color = vec4(v_color, 1.0);
                }
            '''),
        ])

        self.window_size = self.prog.uniforms['WindowSize']

        self.vbo = self.ctx.buffer(struct.pack(
            '15f',
            0.0, 100.0, 1.0, 0.0, 0.0,
            -86.0, -50.0, 0.0, 1.0, 0.0,
            86.0, -50.0, 0.0, 0.0, 1.0,
        ))

        self.vao = self.ctx.simple_vertex_array(self.prog, self.vbo, ['in_vert', 'in_color'])

    def paintGL(self):
        width, height = self.width(), self.height()
        self.ctx.viewport = (0, 0, width, height)
        self.ctx.clear(0.9, 0.9, 0.9)
        self.ctx.enable(ModernGL.BLEND)
        self.window_size.value = (width, height)
        self.vao.render()


app = QtWidgets.QApplication([])
window = QGLControllerWidget()
window.move(QtWidgets.QDesktopWidget().rect().center() - window.rect().center())
window.show()
app.exec_()



import struct

import ModernGL
from kivy.app import App
from kivy.core.window import Window
from kivy.graphics import Callback
from kivy.uix.widget import Widget
from pyrr import Matrix44


class CustomWidget(Widget):
    def __init__(self, **kwargs):
        super(CustomWidget, self).__init__(**kwargs)

        with self.canvas:
            self.ctx = ModernGL.create_context()

            self.prog = self.ctx.program(
                vertex_shader='''
                    #version 330

                    uniform mat4 Mvp;

                    in vec3 in_vert;
                    in vec3 in_color;

                    out vec3 v_color;

                    void main() {
                        v_color = in_color;
                        gl_Position = Mvp * vec4(in_vert, 1.0);
                    }
                '''),
                fragment_shader='''
                    #version 330

                    in vec3 v_color;
                    out vec4 f_color;

                    void main() {
                        f_color = vec4(v_color, 1.0);
                    }
                '''),
            ])

            self.mvp = self.prog.uniforms['Mvp']

            grid = bytearray()

            for i in range(0, 32 + 1):
                grid += struct.pack('6f', i - 16.0, -16.0, 0.0, 0.0, 0.0, 0.0)
                grid += struct.pack('6f', i - 16.0, 16.0, 0.0, 0.0, 0.0, 0.0)
                grid += struct.pack('6f', -16.0, i - 16.0, 0.0, 0.0, 0.0, 0.0)
                grid += struct.pack('6f', 16.0, i - 16.0, 0.0, 0.0, 0.0, 0.0)

            self.vbo = self.ctx.buffer(grid)
            self.vao = self.ctx.simple_vertex_array(self.prog, self.vbo, ['in_vert', 'in_color'])

            Callback(self.draw)

    def draw(self, *args):
        width, height = Window.size
        self.ctx.viewport = (0, 0, width, height)
        self.ctx.clear(0.9, 0.9, 0.9)
        self.ctx.enable(ModernGL.DEPTH_TEST)

        proj = Matrix44.perspective_projection(45.0, width / height, 0.1, 1000.0)
        lookat = Matrix44.look_at(
            (40.0, 30.0, 20.0),
            (0.0, 0.0, 0.0),
            (0.0, 0.0, 1.0),
        )

        self.mvp.write((proj * lookat).astype('float32').tobytes())
        self.vao.render(ModernGL.LINES)

    def ask_update(self, *args):
        self.canvas.ask_update()


class MainApp(App):
    def build(self):
        return CustomWidget()


if __name__ == '__main__':
    MainApp().run()



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import struct

# import ModernGL
# import pyglet

# wnd = pyglet.window.Window(1280, 720)
# ctx = ModernGL.create_context()

# prog = ctx.program(
#     ctx.vertex_shader('''
#         #version 330

#         uniform vec2 WindowSize;

#         in vec2 in_vert;
#         in vec3 in_color;

#         out vec3 v_color;

#         void main() {
#             v_color = in_color;
#             gl_Position = vec4(in_vert / WindowSize * 2.0, 0.0, 1.0);
#         }
#     '''),
#     ctx.fragment_shader('''
#         #version 330

#         in vec3 v_color;
#         out vec4 f_color;

#         void main() {
#             f_color = vec4(v_color, 1.0);
#         }
#     '''),
# ])

# window_size = prog.uniforms['WindowSize']

# vbo = ctx.buffer(struct.pack(
#     '15f',
#     0.0, 100.0, 1.0, 0.0, 0.0,
#     -86.0, -50.0, 0.0, 1.0, 0.0,
#     86.0, -50.0, 0.0, 0.0, 1.0,
# ))

# vao = ctx.simple_vertex_array(prog, vbo, ['in_vert', 'in_color'])


# def update(dt):
#     ctx.viewport = (0, 0, wnd.width, wnd.height)
#     ctx.clear(0.9, 0.9, 0.9)
#     ctx.enable(ModernGL.BLEND)
#     window_size.value = (wnd.width, wnd.height)
#     vao.render()


# pyglet.clock.schedule_interval(update, 1.0 / 60.0)
# pyglet.app.run()



'''
    Renders a blue triangle
'''

import numpy as np

from ported._example import Example


class HelloWorld(Example):
    gl_version = (3, 3)
    title = "Hello World"

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                out vec3 v_color;

                in vec2 in_position;

                vec2 positions[3] = vec2[](
                    vec2(0.0, 0.08),
                    vec2(-0.06, -0.08),
                    vec2(0.06, -0.08)
                );

                vec3 colors[3] = vec3[](
                    vec3(1.0, 0.0, 0.0),
                    vec3(0.0, 1.0, 0.0),
                    vec3(0.0, 0.0, 1.0)
                );

                void main() {
                    gl_Position = vec4(in_position + positions[gl_VertexID], 0.0, 1.0);
                    v_color = colors[gl_VertexID];
                }
            ''',
            fragment_shader='''
                #version 330

                in vec3 v_color;

                layout (location = 0) out vec4 out_color;

                void main() {
                    out_color = vec4(v_color, 1.0);
                }
            ''',
        )

        positions = np.random.uniform(-1.0, 1.0, (100, 2)).astype('f4')

        self.vbo = self.ctx.buffer(positions)
        self.vao = self.ctx.vertex_array(self.prog, [
            (self.vbo, '2f/i', 'in_position'),
        ])
        self.vao.vertices = 3

    def render(self, time, frame_time):
        self.ctx.clear(1.0, 1.0, 1.0)
        self.vao.render(instances=100)


if __name__ == '__main__':
    HelloWorld.run()



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import struct

# import ModernGL
# import pyglet

# wnd = pyglet.window.Window()
# ctx = ModernGL.create_context()

# prog = ctx.program(
#     ctx.vertex_shader('''
#         #version 330

#         in vec2 vert;

#         in vec3 vert_color;
#         out vec3 frag_color;

#         uniform vec2 scale;
#         uniform float rotation;

#         void main() {
#             frag_color = vert_color;
#             mat2 rot = mat2(
#                 cos(rotation), sin(rotation),
#                 -sin(rotation), cos(rotation)
#             );
#             gl_Position = vec4((rot * vert) * scale, 0.0, 1.0);
#         }
#     '''),
#     ctx.fragment_shader('''
#         #version 330

#         in vec3 frag_color;
#         out vec4 color;

#         void main() {
#             color = vec4(frag_color, 1.0);
#         }
#     '''),
# ])

# scale = prog.uniforms['scale']
# rotation = prog.uniforms['rotation']

# vbo = ctx.buffer(struct.pack(
#     '15f',

#     1.0, 0.0,
#     1.0, 0.0, 0.0,

#     -0.5, 0.86,
#     0.0, 1.0, 0.0,

#     -0.5, -0.86,
#     0.0, 0.0, 1.0,
# ))

# vao = ctx.simple_vertex_array(prog, vbo, ['vert', 'vert_color'])


# def update(dt):
#     scale.value = (wnd.height / wnd.width * 0.75, 0.75)
#     ctx.viewport = (0, 0, wnd.width, wnd.height)
#     ctx.clear(0.9, 0.9, 0.9)
#     rotation.value += dt
#     vao.render()


# pyglet.clock.schedule_interval(update, 1.0 / 60.0)
# pyglet.app.run()



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import struct
# import sys

# import ModernGL
# from OpenGL.GLUT import (
#     GLUT_DEPTH, GLUT_DOUBLE, GLUT_ELAPSED_TIME, GLUT_RGB, GLUT_WINDOW_HEIGHT, GLUT_WINDOW_WIDTH, glutCreateWindow,
#     glutDisplayFunc, glutGet, glutIdleFunc, glutInit, glutInitDisplayMode, glutInitWindowSize, glutMainLoop,
#     glutSwapBuffers
# )

# glutInit(sys.argv)
# glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH)
# glutInitWindowSize(800, 600)
# glutCreateWindow(b'Uniforms and Attributes')

# ctx = ModernGL.create_context()

# prog = ctx.program(
#     ctx.vertex_shader('''
#         #version 330

#         in vec2 in_vert;
#         in vec3 in_color;

#         out vec3 v_color;

#         uniform vec2 Scale;
#         uniform float Rotation;

#         void main() {
#             v_color = in_color;
#             mat2 rot = mat2(
#                 cos(Rotation), sin(Rotation),
#                 -sin(Rotation), cos(Rotation)
#             );
#             gl_Position = vec4((rot * in_vert) * Scale, 0.0, 1.0);
#         }
#     '''),
#     ctx.fragment_shader('''
#         #version 330

#         in vec3 v_color;
#         out vec4 f_color;

#         void main() {
#             f_color = vec4(v_color, 1.0);
#         }
#     '''),
# ])

# rotation = prog.uniforms['Rotation']
# scale = prog.uniforms['Scale']

# vbo = ctx.buffer(struct.pack(
#     '15f',

#     1.0, 0.0,
#     1.0, 0.0, 0.0,

#     -0.5, 0.86,
#     0.0, 1.0, 0.0,

#     -0.5, -0.86,
#     0.0, 0.0, 1.0,
# ))

# vao = ctx.simple_vertex_array(prog, vbo, ['in_vert', 'in_color'])


# def display():
#     width, height = glutGet(GLUT_WINDOW_WIDTH), glutGet(GLUT_WINDOW_HEIGHT)
#     elapsed = glutGet(GLUT_ELAPSED_TIME) / 1000.0

#     ctx.clear(0.9, 0.9, 0.9)
#     scale.value = (height / width * 0.75, 0.75)
#     rotation.value = elapsed
#     vao.render()

#     glutSwapBuffers()


# glutDisplayFunc(display)
# glutIdleFunc(display)
# glutMainLoop()



import struct

import moderngl
import pygame
from pygame.locals import DOUBLEBUF, OPENGL

pygame.init()
pygame.display.set_mode((800, 600), DOUBLEBUF | OPENGL)

ctx = moderngl.create_context()

prog = ctx.program(
        vertex_shader='''
    #version 310 es
    in vec2 vert;
    void main() {
        gl_Position = vec4(vert, 0.0, 1.0);
    }
''',
        fragment_shader='''
    #version 310 es
    precision mediump float;
    out vec4 color;

    void main() {
        color = vec4(0.3, 0.5, 1.00, 1.0);
    }
''')


vbo = ctx.buffer(struct.pack('6f', 0.0, 0.8, -0.6, -0.8, 0.6, -0.8))
vao = ctx.simple_vertex_array(prog, vbo, 'vert')

running = True
while running:
    for event in pygame.event.get():
        if event.type == pygame.QUIT:
            running = False

    ctx.clear(0.9, 0.9, 0.9)
    vao.render()

    pygame.display.flip()
    pygame.time.wait(6)



import numpy as np
from pyrr import Matrix44

import moderngl
from moderngl_window.geometry.attributes import AttributeNames
from moderngl_window import geometry

from ported._example import Example


class MugExample(Example):
    title = "Mug"
    gl_version = (3, 3)

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Create a custom attribute name spec
        # so attribute names are not forced to follow gltf standard
        attr_names = AttributeNames(position='in_vert', texcoord_0='in_tex', normal='in_norm')

        # Programs
        self.canvas_prog = self.load_program('mug_mockup/programs/canvas.glsl')
        self.sticker_prog = self.load_program('mug_mockup/programs/sticker.glsl')
        self.mug_prog = self.load_program('mug_mockup/programs/mug.glsl')

        # textures
        self.bg_texture = self.load_texture_2d('mug_mockup/textures/mug-background.jpg')
        self.sticker_texture = self.load_texture_2d('mug_mockup/textures/mug-pymet-logo.png')

        self.canvas_vao = geometry.quad_fs(attr_names=attr_names).instance(self.canvas_prog)

        obj = self.load_scene('mug_mockup/scenes/mug.obj', attr_names=attr_names)
        self.mug_vao = obj.root_nodes[0].mesh.vao.instance(self.mug_prog)

        # Create sticker geometry
        segs = 32
        radius = 29.94
        bottom = 6.601
        top = 57.856
        left = -163.12 * np.pi / 180.0
        right = 11.25 * np.pi / 180.0

        lin = np.linspace(left, right, segs)
        sticker_vertices = np.array([
            np.repeat(np.cos(lin) * radius, 2),
            np.repeat(np.sin(lin) * radius, 2),
            np.tile([bottom, top], segs),
            np.repeat(np.cos(lin), 2),
            np.repeat(np.sin(lin), 2),
            np.tile([0.0, 0.0], segs),
            np.repeat(np.linspace(0.0, 1.0, segs), 2),
            np.tile([0.0, 1.0], segs),
        ])
        self.sticker_vbo = self.ctx.buffer(sticker_vertices.T.astype('f4').tobytes())
        self.sticker_vao = self.ctx.simple_vertex_array(self.sticker_prog, self.sticker_vbo, 'in_vert', 'in_norm', 'in_text')

        # Pre-fill uniforms. These currently do not change during rendering
        proj = Matrix44.perspective_projection(30.0, self.aspect_ratio, 1.0, 1000.0)
        lookat = Matrix44.look_at(
            (46.748, -280.619, 154.391),
            (-23.844, 2.698, 44.493),
            (0.0, 0.0, 1.0),
        )
        mvp = (proj * lookat).astype('f4')
        light = (-143.438, -159.072, 213.268)
        self.mug_prog['Mvp'].write(mvp)
        self.mug_prog['Light'].value = light
        self.sticker_prog['Mvp'].write(mvp)
        self.sticker_prog['Light'].value = light

    def render(self, time, frame_time):
        self.ctx.clear(1.0, 1.0, 1.0)

        # background
        self.ctx.enable_only(moderngl.BLEND)
        self.bg_texture.use()
        self.canvas_vao.render(moderngl.TRIANGLE_STRIP)

        # mug
        self.ctx.enable_only(moderngl.DEPTH_TEST)
        self.mug_vao.render()

        # sticker
        self.ctx.enable_only(moderngl.DEPTH_TEST | moderngl.BLEND)
        self.sticker_texture.use()
        self.sticker_vao.render(moderngl.TRIANGLE_STRIP)


if __name__ == '__main__':
    MugExample.run()



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import math
# import struct

# import GLWindow
# import ModernGL
# import numpy as np


# def perspective(fov, ratio, near, far):
#     zmul = (-2.0 * near * far) / (far - near)
#     ymul = 1.0 / math.tan(fov * math.pi / 360.0)
#     xmul = ymul / ratio
#     return np.matrix([
#         [xmul, 0.0, 0.0, 0.0],
#         [0.0, ymul, 0.0, 0.0],
#         [0.0, 0.0, -1.0, zmul],
#         [0.0, 0.0, -1.0, 0.0],
#     ])


# def lookat(eye, target, up):
#     forward = target - eye
#     forward /= np.linalg.norm(forward)
#     side = np.cross(forward, up)
#     side /= np.linalg.norm(side)
#     upward = np.cross(side, forward)
#     upward /= np.linalg.norm(upward)

#     return np.matrix([
#         [side[0], side[1], side[2], -np.dot(eye, side)],
#         [upward[0], upward[1], upward[2], -np.dot(eye, upward)],
#         [-forward[0], -forward[1], -forward[2], np.dot(eye, forward)],
#         [0.0, 0.0, 0.0, 1.0],
#     ])


# def create_mvp(projection, modelview):
#     return np.transpose(projection @ modelview).astype('float32').tobytes()


# wnd = GLWindow.create_window()
# ctx = ModernGL.create_context()

# prog = ctx.program(
#     ctx.vertex_shader('''
#         #version 330

#         uniform mat4 Mvp;

#         in vec3 in_vert;

#         void main() {
#             gl_Position = Mvp * vec4(in_vert, 1.0);
#         }
#     '''),
#     ctx.fragment_shader('''
#         #version 330

#         out vec4 f_color;

#         void main() {
#             f_color = vec4(0.0, 0.0, 0.0, 1.0);
#         }
#     ''')
# ])

# mvp = prog.uniforms['Mvp']

# grid = bytes()

# for i in range(0, 65):
#     grid += struct.pack('6f', i - 32, -32.0, 0.0, i - 32, 32.0, 0.0)
#     grid += struct.pack('6f', -32.0, i - 32, 0.0, 32.0, i - 32, 0.0)

# vbo = ctx.buffer(grid)
# vao = ctx.simple_vertex_array(prog, vbo, ['in_vert'])

# while wnd.update():
#     ctx.viewport = wnd.viewport
#     ctx.clear(0.9, 0.9, 0.9)

#     proj = perspective(45.0, wnd.size[0] / wnd.size[1], 0.1, 1000.0)
#     view = lookat(
#         np.array([80.0, 70.0, 60.0]),
#         np.array([0.0, 0.0, 0.0]),
#         np.array([0.0, 0.0, 1.0]),
#     )

#     mvp.write(create_mvp(proj, view))

#     vao.render(ModernGL.LINES, 65 * 4)



import GLWindow
import moderngl
from PIL import Image, ImageDraw, ImageFont
import numpy as np

wnd = GLWindow.create_window()
ctx = moderngl.create_context()

prog = ctx.program(
    vertex_shader='''
        #version 330

        in vec2 in_vert;
        in vec3 in_text;

        out vec3 v_text;

        void main() {
            v_text = in_text;
            gl_Position = vec4(in_vert, 0.0, 1.0);
        }
    ''',
    fragment_shader='''
        #version 330

        uniform sampler2DArray Texture;

        in vec3 v_text;

        out vec4 f_color;

        void main() {
            f_color = texture(Texture, v_text);
        }
    ''',
)

x1 = np.cos(np.linspace(0.0, 2.0 * np.pi, 3, False)) * 0.3 * 9/16 - 0.4
y1 = np.sin(np.linspace(0.0, 2.0 * np.pi, 3, False)) * 0.3
z1 = np.full(3, 0.0)

x2 = np.cos(np.linspace(0.0, 2.0 * np.pi, 4, False)) * 0.3 * 9/16
y2 = np.sin(np.linspace(0.0, 2.0 * np.pi, 4, False)) * 0.3
z2 = np.full(4, 1.0)

x3 = np.cos(np.linspace(0.0, 2.0 * np.pi, 5, False)) * 0.3 * 9/16 + 0.4
y3 = np.sin(np.linspace(0.0, 2.0 * np.pi, 5, False)) * 0.3
z3 = np.full(5, 2.0)

shapes = [
    np.dstack([x1, y1, x1, y1, z1]).astype('f4'),
    np.dstack([x2, y2, x2, y2, z2]).astype('f4'),
    np.dstack([x3, y3, x3, y3, z3]).astype('f4'),
]

vbo = ctx.buffer(b''.join(shape.tobytes() for shape in shapes))
dbo = ctx.buffer(np.array([
    3, 1, 0, 0, 0,
    4, 1, 3, 0, 0,
    5, 1, 7, 0, 0,
], dtype='i4'))

images = [
    Image.new('RGB', (256, 256), 'red'),
    Image.new('RGB', (256, 256), 'green'),
    Image.new('RGB', (256, 256), 'blue'),
]

merged = b''.join(img.tobytes() for img in images)

tex = ctx.texture_array((256, 256, 3), 3, merged)
tex.use()

vao = ctx.simple_vertex_array(prog, vbo, 'in_vert', 'in_text')

while wnd.update():
    ctx.clear(1.0, 1.0, 1.0, 1.0)
    # vao.mglo.render(moderngl.TRIANGLE_FAN, 3, 0, 1)
    # vao.mglo.render(moderngl.TRIANGLE_FAN, 4, 3, 1)
    # vao.mglo.render(moderngl.TRIANGLE_FAN, 5, 7, 1)
    vao.render_indirect(dbo, moderngl.TRIANGLE_FAN)



from pyrr import Matrix44

import _example


class Example(_example.Example):
    title = "Simple 3D Example"

    STATIC_COLOR_MODE = 0
    STATIC_COLOR_WITH_LIGHT_MODE = 1
    ATTRIB_COLOR_WITH_LIGHT_MODE = 2
    TEXTURE_WITH_LIGHT_MODE = 3

    def set_camera(self, fov, eye, target):
        proj = Matrix44.perspective_projection(fov, self.aspect_ratio, 0.1, 1000.0)
        look = Matrix44.look_at(eye, target, (0.0, 0.0, 1.0))
        self.prog['Mvp'].write((proj * look).astype('f4').tobytes())
        self.prog['Light'].value = eye

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.prog = self.ctx.program(
            vertex_shader='''
                #version 330

                uniform mat4 Mvp;

                in vec3 in_vert;
                in vec3 in_norm;
                in vec3 in_color;
                in vec2 in_text;

                out vec3 v_vert;
                out vec3 v_norm;
                out vec3 v_color;
                out vec2 v_text;

                void main() {
                    gl_Position = Mvp * vec4(in_vert, 1.0);
                    v_vert = in_vert;
                    v_norm = in_norm;
                    v_color = in_color;
                    v_text = in_text;
                }
            ''',
            fragment_shader='''
                #version 330

                uniform sampler2D Texture;
                uniform int RenderMode;
                uniform vec3 Color;
                uniform vec3 Light;

                in vec3 v_vert;
                in vec3 v_norm;
                in vec3 v_color;
                in vec2 v_text;

                out vec4 f_color;

                void main() {
                    float lum = 0.2 + 0.8 * abs(dot(normalize(Light - v_vert), normalize(v_norm)));
                    if (RenderMode == 0) {
                        f_color = vec4(Color, 1.0);
                    } else if (RenderMode == 1) {
                        f_color = vec4(Color * lum, 1.0);
                    } else if (RenderMode == 2) {
                        f_color = vec4(v_color * lum, 1.0);
                    } else if (RenderMode == 3) {
                        f_color = texture(Texture, v_text) * vec4(lum, lum, lum, 1.0);
                    }
                }
            ''',
        )



"""
NOTE: This example is from ModernGL 4 or earlier. We simply disable and archive them for now.
"""

# import struct
# from math import cos, sin
# from random import randrange

# import GLWindow
# import ModernGL
# from pyrr import Matrix44


# def create_brick_texture(size):
#     result = bytearray()

#     for z in range(size[2]):
#         for y in range(size[1]):
#             for x in range(size[0]):

#                 if x != size[0] - 1 and y != size[1] - 1 and z != size[2] - 1:
#                     n1, n2, n3 = randrange(0, 16), randrange(0, 16), randrange(0, 16)
#                     result += struct.pack('BBB', 220 + n1, 31 + n2, 0 + n3)

#                 else:
#                     n = randrange(0, 16)
#                     result += struct.pack('BBB', 240 - n, 240 - n, 240 - n)

#     return bytes(result)


# wnd = GLWindow.create_window()
# ctx = ModernGL.create_context()

# prog = ctx.program(
#     ctx.vertex_shader('''
#         #version 330

#         uniform mat4 Mvp;
#         uniform vec3 TextureSize;

#         in vec3 vert;
#         out vec3 v_text;

#         void main() {
#             v_text = vert * 100.0 / TextureSize;
#             gl_Position = Mvp * vec4(vert, 1.0);
#         }
#     '''),
#     ctx.fragment_shader('''
#         #version 330

#         uniform sampler3D Texture;

#         in vec3 v_text;
#         out vec4 color;

#         void main() {
#             color = texture(Texture, v_text);
#         }
#     '''),
# ])

# mvp = prog.uniforms['Mvp']
# texture_size = prog.uniforms['TextureSize']

# vbo = ctx.buffer(struct.pack(
#     '36f',

#     0.0, 0.0, 0.0,
#     2.0, 0.0, 0.0,
#     0.0, 2.0, 0.0,

#     0.0, 0.0, 0.0,
#     2.0, 0.0, 0.0,
#     0.0, 0.0, 2.0,

#     0.0, 0.0, 0.0,
#     0.0, 2.0, 0.0,
#     0.0, 0.0, 2.0,

#     2.0, 0.0, 0.0,
#     0.0, 2.0, 0.0,
#     0.0, 0.0, 2.0,
# ))

# vao = ctx.simple_vertex_array(prog, vbo, ['vert'])

# size = (20, 16, 8)
# texture_size.value = size
# texure = ctx.texture3d(size, 3, create_brick_texture(size))
# texure.filter = ModernGL.NEAREST
# texure.use()

# ctx.enable(ModernGL.DEPTH_TEST)

# while wnd.update():
#     angle = wnd.time
#     width, height = wnd.size
#     proj = Matrix44.perspective_projection(45.0, width / height, 0.1, 1000.0)
#     lookat = Matrix44.look_at(
#         (cos(angle) * 5.0, sin(angle) * 5.0, 2.5),
#         (0.0, 0.0, 0.5),
#         (0.0, 0.0, 1.0),
#     )

#     mvp.write((proj * lookat).astype('float32').tobytes())

#     ctx.viewport = wnd.viewport
#     ctx.clear(0.9, 0.9, 0.9)
#     vao.render()
